"""Module doc extracts source code documentation from a KCL AST
"""
import io
import os

from typing import Union, List, Dict, Optional
from pathlib import Path

import kclvm.kcl.error as kcl_error
import kclvm.compiler.vfs as vfs
import kclvm.compiler.parser.parser as parser
import kclvm.tools.docs.formats as doc_formats
import kclvm.kcl.types.checker as type_checker
import kclvm.api.object as obj_pkg
import kclvm.kcl.ast as ast
import kclvm.tools.docs.model_pb2 as model
import kclvm.tools.docs.writer as writer
import kclvm.tools.docs.reader as reader
import kclvm.tools.docs.link_resolver as link_resolver
import kclvm.tools.docs.doc_escaper as doc_escaper
import kclvm.tools.docs.doc_parser as doc_parser
import kclvm.tools.docs.i18n as i18n
import kclvm.tools.docs.utils as utils

# ---------------------------------------------------
# Constants
# ---------------------------------------------------

KCL_SUFFIX = "*.k"
INVALID_OUTPUT_FORMAT_MSG = "invalid output format, expected: yaml, json or markdown"
INVALID_I18N_FORMAT_MSG = "invalid i18n metadata format, expected: yaml, json"

# ---------------------------------------------------
# User interface functions used by kcl-doc cli
# ---------------------------------------------------


def kcl_doc_generate(
    kcl_files: List[str],
    output: str,
    format: str = doc_formats.KCLDocFormat.MARKDOWN,
    locale: str = "en",
    recursively=False,
    repo_url: str = None,
    i18n_path: str = None,
    with_locale_suffix: bool = False,
) -> None:
    """
    generate a displayable doc file of a kcl file
    :param kcl_files: the kcl file paths to generate doc on
    :param output: the dir path to output the generated doc file
    :param format: the document format to generate
    :param locale: the document locale to generate
    :param recursively: if search for the kcl files to generate doc on recursively
    :param repo_url: the url to the source code repo
    :param i18n_path: the i18n input path
    """
    # check if the format and locale is valid
    locale = locale.lower()
    format_upper = format.upper()
    if format_upper not in doc_formats.KCLDocFormat.MAPPING:
        kcl_error.report_exception(
            err_type=kcl_error.ErrType.CompileError_TYPE,
            arg_msg=INVALID_OUTPUT_FORMAT_MSG,
        )
    i18n.check_locale(locale)
    # check if all the files exist
    if len(kcl_files) == 0:
        kcl_error.report_exception(
            err_type=kcl_error.ErrType.CannotFindModule_TYPE,
            arg_msg="Empty list of input KCL files",
        )
    for kcl_file in kcl_files:
        if not Path(kcl_file).exists():
            kcl_error.report_exception(
                err_type=kcl_error.ErrType.CannotFindModule_TYPE,
                arg_msg=f"Cannot find the kcl file, please check whether the file path {kcl_file} exists"
                if kcl_file
                else f"The file path {kcl_file} is None",
            )
    # parse module docs
    module_docs = _parse_file_docs(
        kcl_files, format_upper, locale, recursively, repo_url, with_locale_suffix
    )
    # write to output
    path = Path(output).resolve()
    for _, doc in module_docs.items():
        # if the i18n file exists, use the doc content decoded from the existing file
        i18n_file_path, i18n_format = _existed_i18n_file_path(
            i18n_path, path, doc.name, locale
        )
        if i18n_file_path and i18n_format:
            doc = _read_from_file(i18n_format, i18n_file_path)
            _escape_text({doc.name: doc}, format_upper)
        doc_file_path = _doc_file_path(
            path / Path(doc.relative_path).parent,
            doc.name,
            locale,
            format_upper,
            with_locale_suffix,
        )
        _write_to_file(doc, format_upper, doc_file_path)


def kcl_i18n_init(
    kcl_files: List[str],
    output: str,
    format: str,
    locale: str = "en",
    recursively=False,
    with_locale_suffix: bool = False,
) -> None:
    """
    init an i18n source doc file of a KCL file. Users can then modify the document part of it and generate docs in other formats based on it
    :param kcl_files: the kcl file paths to generate doc on
    :param output: the dir path to output the inited i18n file
    :param format: the document format to init
    :param locale: the document locale to init
    :param recursively: if search for the kcl files to generate doc on recursively
    """
    # check if the format and locale is valid
    locale = locale.lower()
    format_upper = format.upper()
    if format_upper not in doc_formats.KCLI18NFormat.MAPPING:
        kcl_error.report_exception(
            err_type=kcl_error.ErrType.CompileError_TYPE,
            arg_msg=INVALID_I18N_FORMAT_MSG,
        )
    i18n.check_locale(locale)
    # check if all the files exist
    if len(kcl_files) == 0:
        kcl_error.report_exception(
            err_type=kcl_error.ErrType.CannotFindModule_TYPE,
            arg_msg="Empty list of input KCL files",
        )
    for kcl_file in kcl_files:
        if not Path(kcl_file).exists():
            kcl_error.report_exception(
                err_type=kcl_error.ErrType.CannotFindModule_TYPE,
                arg_msg=f"Cannot find the kcl file, please check whether the file path {kcl_file} exists"
                if kcl_file
                else f"The file path {kcl_file} is None",
            )
    # parse module docs
    module_docs = _parse_file_docs(
        kcl_files,
        format_upper,
        locale,
        recursively,
        with_locale_suffix=with_locale_suffix,
    )
    # write to output
    path = Path(output).resolve()
    path.mkdir(parents=True, exist_ok=True)
    for _, doc in module_docs.items():
        i18n_path = _i18n_file_path(path, doc.name, locale, format_upper)
        # reset doc content before gen i18n file
        _write_to_file(doc, format_upper, i18n_path)


def kcl_i18n_info(kcl_files: List[str], format: str, locale: str) -> None:
    """
    show an i18n source doc of a kcl file
    :param kcl_files: the kcl files to show i18n info
    :param format: the i18n file format to display
    :param locale: the i18n file locale to display
    """
    # check if the format and locale is valid
    locale = locale.lower()
    format_upper = format.upper()
    if format_upper not in doc_formats.KCLI18NFormat.MAPPING:
        kcl_error.report_exception(
            err_type=kcl_error.ErrType.CompileError_TYPE,
            arg_msg=INVALID_I18N_FORMAT_MSG,
        )
    i18n.check_locale(locale)

    for kcl_file in kcl_files:
        # calculate the i18n file path
        path = _i18n_file_path(
            Path(kcl_file).parent, utils.module_name(kcl_file), locale, format_upper
        )
        # display the doc content read from i18n file to console
        print(f"i18n file: {path} \ncontent:")
        path = Path(path)
        if path.exists():
            module_doc = _read_from_file(format_upper, path)
            if module_doc:
                output = io.StringIO()
                _write_to_io(module_doc, format_upper, output)
                print(output.getvalue())
            else:
                print("Failed to parse i18n locale file, get nothing.")
        else:
            print(
                f"I18n local file not exist. \n    KCL file path: {kcl_file}, format: {format}, locale: {locale}"
            )


# ---------------------------------------------------
# Internal functions
# ---------------------------------------------------


def _read_from_file(format: str, in_file: Union[str, Path]) -> model.ModuleDoc:
    with Path(in_file).open(mode="r", encoding="utf-8") as in_io:
        module_doc = _read_from_io(format, in_io)
    return module_doc


def _read_from_io(format: str, in_io: io.TextIOBase) -> model.ModuleDoc:
    doc_reader = reader.factory.get(format, in_io)
    return doc_reader.read_doc()


def _write_to_file(
    doc: model.ModuleDoc, format: str, out_file: Union[str, Path]
) -> None:
    Path(out_file).parent.mkdir(parents=True, exist_ok=True)
    with Path(out_file).open(mode="w", encoding="utf-8") as out:
        _write_to_io(doc, format, out)


def _write_to_io(doc: model.ModuleDoc, format: str, out: io.TextIOBase) -> None:
    doc_writer = writer.factory.get(format, out)
    doc_writer.write_doc(doc)


def _doc_file_name(
    module_name: str, locale: str, format: str, with_locale_suffix: bool = False
) -> str:
    return (
        f"doc_{module_name}_{locale}" + doc_formats.KCLDocSuffix.TO_SUFFIX[format]
        if with_locale_suffix
        else f"doc_{module_name}{doc_formats.KCLDocSuffix.TO_SUFFIX[format]}"
    )


def _doc_file_path(
    path: Path, name: str, locale: str, format: str, with_locale_suffix: bool = False
) -> Path:
    return path / _doc_file_name(
        module_name=name,
        locale=locale,
        format=format,
        with_locale_suffix=with_locale_suffix,
    )


def _i18n_file_path(path: Path, name: str, locale: str, format: str) -> Path:
    return path / (f"i18n_{name}_{locale}" + doc_formats.KCLDocSuffix.TO_SUFFIX[format])


def _existed_i18n_file_path(
    i18n_path: str, path: Path, name: str, locale: str
) -> (Optional[Path], Optional[str]):
    """
    check if there are existed i18n files for current source file path.
    When there exists more than one possible i18n files, the priority to return is: KCL -> YAML -> JSON

    :param i18n_path: the i18n input file path
    :param path: the dir path that contains KCL source file
    :param name: the KCL module name to generate doc on
    :param locale: the target locale
    :return: the i18n file path if existed, else return None.
    """
    if i18n_path:
        i18n_path = Path(i18n_path)
        if not i18n_path.exists():
            return None, None
        if (
            i18n_path.is_file()
            and i18n_path.suffix in doc_formats.KCLI18NFormat.FROM_SUFFIX
        ):
            return i18n_path, doc_formats.KCLI18NFormat.FROM_SUFFIX[i18n_path.suffix]
        elif i18n_path.is_dir():
            path = i18n_path
        else:
            return None, None
    yaml_i18n_path = _i18n_file_path(path, name, locale, doc_formats.KCLI18NFormat.YAML)
    if yaml_i18n_path.exists():
        return yaml_i18n_path, doc_formats.KCLI18NFormat.YAML

    json_i18n_path = _i18n_file_path(path, name, locale, doc_formats.KCLI18NFormat.JSON)
    if json_i18n_path.exists():
        return json_i18n_path, doc_formats.KCLI18NFormat.JSON

    return None, None


def _escape_text(module_docs: Dict[str, model.ModuleDoc], format: str):
    escaper = doc_escaper.factory.get(format)
    for path, module in module_docs.items():
        module_docs[path] = escaper.escape(module)


def _resolve_link(
    root: str,
    module_docs: Dict[str, model.ModuleDoc],
    format: str,
    locale: str,
    with_locale_suffix: bool = False,
):
    resolver = link_resolver.factory.get(format, _doc_file_name)
    for path, module in module_docs.items():
        module_docs[path] = resolver.resolve(module, root, locale, with_locale_suffix)


def _parse_file_docs(
    kcl_files: List[str],
    format: str,
    locale: str,
    recursively: bool,
    repo_url: str = None,
    with_locale_suffix: bool = False,
) -> Dict[str, model.ModuleDoc]:
    pkgs: List[str] = []
    if recursively:
        for kcl_file in kcl_files:
            pkgs.extend(_find_kcl_pkgs_recursively(Path(kcl_file)))
        pkgs = list(set(pkgs))
    else:
        pkgs = kcl_files
    if len(pkgs) == 0:
        return {}
    root: str = vfs.GetPkgRoot(pkgs[0]) or os.path.dirname(pkgs[0])
    module_docs: Dict[str, model.ModuleDoc] = {}
    trigger_lines = [
        "import .{}\n".format(
            str(Path(pkg).relative_to(root))
            .replace("/", ".")  # Unix path separator
            .replace("\\", ".")  # Windows path separator
            .rstrip(".k")
        )
        for pkg in pkgs
    ]
    trigger_file = Path(root) / "trigger_doc_gen.k"
    with open(trigger_file, "w") as f:
        f.writelines(trigger_lines)
    try:
        prog = parser.LoadProgram(
            str(trigger_file),
            work_dir=root,
        )
        type_checker.ResolveProgramImport(prog)
        checker = type_checker.TypeChecker(prog, type_checker.CheckConfig())
        checker.check_import(prog.MAIN_PKGPATH)
        checker.init_global_types()
        del prog.pkgs[prog.MAIN_PKGPATH]
    finally:
        os.remove(trigger_file)
    pkg_docs = _parse_program_docs(prog, checker, repo_url)
    _escape_text(pkg_docs, format)
    _resolve_link(prog.root, pkg_docs, format, locale, with_locale_suffix)

    module_docs.update(pkg_docs)
    return module_docs


def _parse_program_docs(
    program: ast.Program, checker: type_checker.TypeChecker, repo_url: str = None
) -> Dict[str, model.ModuleDoc]:
    pkgs: Dict[str, List[ast.Module]] = program.pkgs
    module_docs: Dict[str, model.ModuleDoc] = {}
    for pkgpath, modules in pkgs.items():
        for m in modules:
            schema_docs: list = []
            schema_list = m.GetSchemaList()
            # only generate module doc if the module contains schemas
            if schema_list:
                for schema in schema_list:
                    current_scope = checker.scope_map[pkgpath]
                    schema_obj_type = current_scope.elems[schema.name].type
                    if isinstance(schema_obj_type, obj_pkg.KCLSchemaDefTypeObject):
                        schema_doc = doc_parser.SchemaDocParser(
                            schema=schema,
                            schema_type=schema_obj_type.schema_type,
                            root=program.root,
                        ).doc
                        schema_docs.append(schema_doc)
                module_doc = model.ModuleDoc(
                    name=utils.module_name(m.filename),
                    relative_path=m.relative_filename,
                    doc=m.doc,
                    schemas=schema_docs,
                    source_code_url=repo_url,
                )
                module_docs[m.relative_filename] = module_doc
    return module_docs


def _find_kcl_pkgs_recursively(file: Path) -> List[str]:
    if file.is_file():
        return [str(file)]
    all_files = file.rglob("*.k")
    pkgs: List[str] = [str(f.parent) for f in all_files]
    return list(set(pkgs))



import base64 as _base64


def KMANGLED_encode(value: str, encoding: str = "utf-8") -> str:
    return _base64.b64encode(value.encode(encoding)).decode(encoding)


def KMANGLED_decode(value: str, encoding: str = "utf-8") -> str:
    return _base64.b64decode(value).decode(encoding)



from .union import union, merge
from .common import handle_subscript
from .lazy import ValueCache, Backtracking, SchemaEvalContext

__all__ = [
    "union",
    "merge",
    "handle_subscript",
    "ValueCache",
    "SchemaEvalContext",
    "Backtracking",
]



# Copyright 2020 The KCL Authors. All rights reserved.

import typing
import pathlib

from lark import Lark
from lark.indenter import Indenter
from lark.tree import Tree as LarkTree
from lark.lexer import Token as LarkToken

import kclvm.kcl.error as kcl_error
import kclvm.compiler.parser.lark_pb2 as lark_pb

_CACHE_FILE = pathlib.Path(__file__).parent.joinpath("lark_parser.pickle")
START_RULE = "start"

filename = ""


class KCLIndenter(Indenter):
    NL_type = "NEWLINE"
    OPEN_PAREN_types = ["LPAR", "LSQB", "LBRACE"]
    CLOSE_PAREN_types = ["RPAR", "RSQB", "RBRACE"]
    INDENT_type = "_INDENT"
    DEDENT_type = "_DEDENT"
    tab_len = 4
    _indent_has_space = False  # Mark whether there are spaces in an indent_level
    _indent_has_tab = False  # Mark whether there are tabs in an indent_level

    def __init__(self):
        super().__init__()
        self.reset_indent_space_tab()

    def reset_indent_space_tab(self):
        self._indent_has_space = False
        self._indent_has_tab = False

    def process(self, stream):
        self.paren_level = 0
        self.indent_level = [0]
        self.reset_indent_space_tab()
        return self._process(stream)

    def check_tab_error(self, space_count, tab_count, line=None, column=None):
        """Check TabError: Inconsistent use of tabs and spaces in indentation"""
        self._indent_has_space = True if space_count else self._indent_has_space
        self._indent_has_tab = True if tab_count else self._indent_has_tab
        if self._indent_has_space and self._indent_has_tab:
            import kclvm.compiler.parser.lark_parser as lark_parser

            kcl_error.report_exception(
                err_type=kcl_error.ErrType.TabError_TYPE,
                file_msgs=[
                    kcl_error.ErrFileMsg(
                        filename=lark_parser.filename, line_no=line, col_no=column
                    )
                ],
            )

    def handle_NL(self, token):
        """Do not edit it, inherit from base class 'Indenter'"""
        if self.paren_level > 0:
            return

        yield token

        indent_str = token.rsplit("\n", 1)[1]  # Tabs and spaces
        space_count, tab_count = indent_str.count(" "), indent_str.count("\t")
        indent = space_count + tab_count * self.tab_len
        self.check_tab_error(space_count, tab_count, token.end_line, token.end_column)

        if indent > self.indent_level[-1]:
            self.indent_level.append(indent)
            yield LarkToken.new_borrow_pos(self.INDENT_type, indent_str, token)
        else:
            while indent < self.indent_level[-1]:
                self.indent_level.pop()
                self.reset_indent_space_tab()
                lark_token = LarkToken.new_borrow_pos(
                    self.DEDENT_type, indent_str, token
                )
                yield lark_token

            if indent != self.indent_level[-1]:
                import kclvm.compiler.parser.lark_parser as lark_parser

                kcl_error.report_exception(
                    err_type=kcl_error.ErrType.IndentationError_TYPE,
                    file_msgs=[
                        kcl_error.ErrFileMsg(
                            filename=lark_parser.filename,
                            line_no=lark_token.end_line,
                            col_no=lark_token.end_column,
                        )
                    ],
                    arg_msg=kcl_error.INDENTATION_ERROR_MSG.format(str(indent)),
                )


_kcl_lark_parser: typing.Optional[Lark] = None


def GetKclLarkParser() -> Lark:
    global _kcl_lark_parser

    if _kcl_lark_parser is None:
        _kcl_lark_parser = Lark.open(
            "../../kcl/grammar/kcl.lark",
            parser="lalr",
            propagate_positions=True,
            postlex=KCLIndenter(),
            rel_to=__file__,
            cache=str(_CACHE_FILE),
        )

    return _kcl_lark_parser


def IsRuleType(node_type: str) -> bool:
    return node_type.islower()


def IsTokenType(node_type: str) -> bool:
    return node_type.isupper()


def GetNode(
    node: lark_pb.Tree, node_type: str, *more_node_type: str
) -> typing.Optional[lark_pb.Tree]:
    node_list = GetNodeList(node, node_type, *more_node_type, max_size=1)
    return node_list[0] if node_list else None


def GetNodeList(
    node: lark_pb.Tree,
    target_node_type: str,
    *more_target_node_type: str,
    max_size=0,
    recursively: bool = True
) -> typing.List[lark_pb.Tree]:
    node_type_list = [target_node_type, *more_target_node_type]

    if not node:
        return []

    if node.type in node_type_list:
        return [node]  # OK

    # try sub node
    node_list = []
    for n in node.children or []:
        if n.type in node_type_list:
            node_list.append(n)
            if 0 < max_size <= len(node_list):
                return node_list
            continue
        if recursively:
            node_list.extend(
                GetNodeList(
                    n, target_node_type, *more_target_node_type, max_size=max_size
                )
            )
        if 0 < max_size <= len(node_list):
            return node_list

    return node_list


def WalkTree(t: lark_pb.Tree, walk_fn):
    walk_fn(t)
    for n in t.children:
        WalkTree(n, walk_fn)


def ParseFile(filename: str, code: str, ignore_file_line: bool = False) -> lark_pb.Tree:
    if not code:
        with open(filename) as f:
            code = str(f.read())
    return ParseCode(code, ignore_file_line=ignore_file_line)


def ParseCode(src: str, ignore_file_line: bool = False) -> lark_pb.Tree:
    def _pb_build_Tree(_t: LarkTree) -> lark_pb.Tree:
        if isinstance(_t, LarkTree):
            rule_type = _t.data

            assert rule_type.islower()
            assert len(_t.children) >= 0

            # Empty file and return a empty lark tree node
            if rule_type == START_RULE and not _t.children:
                t = lark_pb.Tree(
                    type=rule_type,
                    token_value="",
                    children=[],
                )
            elif not ignore_file_line:
                t = lark_pb.Tree(
                    type=rule_type,
                    token_value="",  # rule, not token
                    children=[],
                    line=_t.meta.line,
                    column=_t.meta.column,
                    end_line=_t.meta.end_line,
                    end_column=_t.meta.end_column,
                )
            else:
                t = lark_pb.Tree(
                    type=rule_type, token_value="", children=[]  # rule, not token
                )

            for v in _t.children:
                t.children.append(_pb_build_Tree(v))

            return t

        if isinstance(_t, LarkToken):
            token_type = _t.type

            assert token_type.isupper()

            if not ignore_file_line:
                return lark_pb.Tree(
                    type=token_type,
                    token_value=_t.value,
                    children=[],
                    line=_t.line,
                    column=_t.column,
                    end_line=_t.end_line,
                    end_column=_t.end_column,
                )
            else:
                return lark_pb.Tree(type=token_type, token_value=_t.value, children=[])

        return lark_pb.Tree()

    # To prevent empty files and files that only contain line continuation symbols
    src += "\n"
    tree = GetKclLarkParser().parse(src)
    return _pb_build_Tree(tree)



# Copyright 2021 The KCL Authors. All rights reserved.

from typing import cast, List, Union
from io import StringIO

import kclvm.kcl.ast as ast
import kclvm.kcl.types as types
import kclvm.api.object as objpkg
import kclvm.tools.printer as printer
import kclvm.compiler.parser.parser as parser

from .parser import (
    parse_code,
    Proto,
    Message,
    Field,
    Map,
    Enum,
    EnumField,
    Type,
    KeyType,
    OneOf,
    OneOfField,
)
from .types import (
    PROTOBUF_SCALAR_TYPE_TO_KCL_MAPPING,
    KCL_SCALAR_TYPE_TO_PROTOBUF_TYPE_MAPPING,
    KCL_SCALAR_TYPE_TO_PROTOBUF_KEY_TYPE_MAPPING,
)
from .printer import print_node_to_string


CODE_INPUT = "<input>"
PROTO3_SYNTAX = "proto3"


def protobuf_to_kcl(proto_code: str) -> str:
    """Covert a protobuf code string to a KCL code string.

    Parameters
    ----------
    proto_code : str. The protobuf code string.

    Returns
    -------
    kcl_code : str. The KCL code string.
    """
    kcl_module = protobuf_to_kcl_ast(proto_code)
    buf = StringIO()
    printer.PrintAST(kcl_module, buf)
    return buf.getvalue()


def kcl_to_protobuf(kcl_code: str) -> str:
    """Covert a a KCL code string to a protobuf code string.

    Parameters
    ----------
    kcl_code : str. The KCL code string.

    Returns
    -------
    proto_code : str. The protobuf code string.
    """
    ast_program = parser.LoadProgram(CODE_INPUT, k_code_list=[kcl_code])
    types.ResolveProgram(ast_program)
    main_ast = ast_program.pkgs[ast.Program.MAIN_PKGPATH][0]
    proto = Proto(
        syntax=PROTO3_SYNTAX,
        statements=[
            convert_schema_to_message(schema) for schema in main_ast.GetSchemaList()
        ],
    )
    return print_node_to_string(proto)


def protobuf_to_kcl_ast(proto_code: str) -> ast.Module:
    """Covert a protobuf code string to a KCL code string.

    Parameters
    ----------
    proto_code : str. The protobuf code string.

    Returns
    -------
    kcl_ast : ast.Module. The KCL Module AST node.
    """
    proto = parse_code(proto_code)
    kcl_module = ast.Module(line=1, column=1)
    for statement in proto.statements:
        if isinstance(statement, Message):
            message = cast(Message, statement)
            kcl_module.body.extend(convert_message_to_schema_list(message))
        elif isinstance(statement, Enum):
            enum = cast(Enum, statement)
            kcl_module.body.append(convert_enum_to_type_alias(enum))
        # TODO: Import node on multi proto files
    return kcl_module


def convert_proto_type_to_kcl_type(tpe: Union[Type, str]) -> str:
    """Convert a proto type to a KCL type"""
    return (
        PROTOBUF_SCALAR_TYPE_TO_KCL_MAPPING.get(tpe.value)
        if isinstance(tpe, (Type, KeyType))
        else (tpe or "any")
    )


def convert_kcl_type_to_proto_type(
    tpe: types.Type, is_proto_key_type: bool = False
) -> Union[str, KeyType, Type]:
    """Convert a kcl type to a proto type"""
    if isinstance(tpe, objpkg.KCLNamedTypeObject):
        return tpe.name
    elif isinstance(
        tpe,
        (
            objpkg.KCLIntTypeObject,
            objpkg.KCLFloatTypeObject,
            objpkg.KCLStringTypeObject,
            objpkg.KCLBoolTypeObject,
        ),
    ):
        return (
            KCL_SCALAR_TYPE_TO_PROTOBUF_KEY_TYPE_MAPPING.get(tpe.type_str())
            if is_proto_key_type
            else KCL_SCALAR_TYPE_TO_PROTOBUF_TYPE_MAPPING.get(tpe.type_str())
        )
    raise ValueError(f"Unsupported KCL type {tpe} to proto type")


def convert_message_to_schema_list(message: Message) -> List[ast.Stmt]:
    """Convert proto Message to KCL schema list because a
    proto Message can be nested definition.
    """
    if not message or not isinstance(message, Message):
        raise ValueError(f"Invalid parameter message {message}, expected Message")
    schema = ast.SchemaStmt()
    # Mapping the message name to the schema name
    schema.name = message.name
    results = []
    results.append(schema)
    for node in message.body:
        schema_attr = ast.SchemaAttr()
        if isinstance(node, Field):
            field = cast(Field, node)
            schema_attr.name = field.name
            kcl_type_str = convert_proto_type_to_kcl_type(field.type)
            schema_attr.type_str = (
                f"[{kcl_type_str}]" if field.repeated else kcl_type_str
            )
            schema.body.append(schema_attr)
        elif isinstance(node, Map):
            field = cast(Map, node)
            schema_attr.name = field.name
            kcl_key_type_str = convert_proto_type_to_kcl_type(field.key_type)
            kcl_value_type_str = convert_proto_type_to_kcl_type(field.type)
            schema_attr.type_str = (
                "{" + kcl_key_type_str + ":" + kcl_value_type_str + "}"
            )
            schema.body.append(schema_attr)
        elif isinstance(node, OneOf):
            field = cast(OneOf, node)
            schema_attr.name = field.name
            # OneOfField
            schema_attr.type_str = " | ".join(
                [convert_proto_type_to_kcl_type(field.type) for field in field.fields]
            )
            schema.body.append(schema_attr)
        elif isinstance(node, Message):
            results.extend(convert_message_to_schema_list(node))
    return results


def convert_enum_to_type_alias(enum: Enum) -> ast.TypeAliasStmt:
    """Convert a proto Enum to a KCL TypeAliasStmt"""
    if not enum or not isinstance(enum, Enum):
        raise ValueError(f"Invalid parameter Enum {enum}, expected Enum")
    type_alias = ast.TypeAliasStmt()
    type_alias.type_name = ast.ASTFactory.get_ast_identifier(enum.name)
    values = [str(node.value) for node in enum.body if isinstance(node, EnumField)]
    type_alias.type_value = ast.Type()
    type_alias.type_value.plain_type_str = " | ".join(values)
    return type_alias


def convert_schema_to_message(schema: ast.SchemaStmt) -> Message:
    """Convert KCL schema to a proto Message"""
    if not schema or not isinstance(schema, ast.SchemaStmt):
        raise ValueError(f"Invalid parameter schema {schema}, expected ast.SchemaStmt")
    return Message(
        name=schema.name,
        body=[
            convert_schema_attr_to_message_body(schema_attr, i + 1)
            for i, schema_attr in enumerate(schema.body)
            if isinstance(schema_attr, ast.SchemaAttr)
        ],
    )


def convert_schema_attr_to_message_body(
    schema_attr: ast.SchemaAttr,
    number: int = 1,
) -> Union[Field, Map, OneOf]:
    """Convert KCL schema attr to a proto Message body item"""
    if not schema_attr or not isinstance(schema_attr, ast.SchemaAttr):
        raise ValueError(
            f"Invalid parameter schema {schema_attr}, expected ast.SchemaAttr"
        )
    type_node = types.parse_type_str(schema_attr.type_str)
    if isinstance(type_node, objpkg.KCLDictTypeObject):
        field = Map(
            name=schema_attr.name,
            key_type=convert_kcl_type_to_proto_type(type_node.key_type, True),
            type=convert_kcl_type_to_proto_type(type_node.value_type),
            number=number,
            options=[],
        )
    elif isinstance(type_node, objpkg.KCLUnionTypeObject):
        # Convert union type to one of
        field = OneOf(
            name=schema_attr.name,
            fields=[],
        )
        filed_name_prefix = schema_attr.name
        for i, tpe in enumerate(type_node.types):
            field.fields.append(
                OneOfField(
                    name=filed_name_prefix + f"{i + 1}",
                    number=i + 1,
                    type=types.type_to_kcl_type_annotation_str(tpe),
                    options=[],
                )
            )
    else:
        field = Field(
            name=schema_attr.name,
            type="",
            repeated=False,
            options=[],
            number=number,
        )
        if isinstance(type_node, objpkg.KCLListTypeObject):
            field.repeated = True
            field.type = convert_kcl_type_to_proto_type(type_node.item_type)
        else:
            field.type = convert_kcl_type_to_proto_type(type_node)
    return field



# Copyright 2020 The KCL Authors. All rights reserved.

import argparse
import sys
import json

import kclvm.compiler.extension.plugin.plugin as plugin


def Main():
    parser = argparse.ArgumentParser(prog="kcl-plugin")

    subparsers = parser.add_subparsers(
        dest="kcl_plugin_subcmd_name", help="kcl plugin sub commands"
    )

    parser_list = subparsers.add_parser("list", help="list all plugins")

    parser_init = subparsers.add_parser("init", help="init a new plugin")

    parser_info = subparsers.add_parser("info", help="show plugin document")
    subparsers.add_parser("gendoc", help="gen all plugins document")
    subparsers.add_parser("version", help="show plugin version")
    parser_test = subparsers.add_parser("test", help="test plugin")

    parser_list.add_argument(dest="plugin_keyword", metavar="keyword", nargs="?")

    parser_init.add_argument(dest="plugin_name", metavar="name")
    parser_info.add_argument(dest="plugin_name", metavar="name", nargs="?")

    parser_test.add_argument(dest="plugin_name", metavar="name")

    args = parser.parse_args()

    if args.kcl_plugin_subcmd_name == "list":
        plugin_root = plugin.get_plugin_root()
        if not plugin_root:
            print("# plugin_root: <not found>")
            sys.exit(0)

        print(f"# plugin_root: {plugin.get_plugin_root()}")
        names = plugin.get_plugin_names()

        count = 0
        for name in names:
            if args.plugin_keyword and args.plugin_keyword not in name:
                continue
            info = plugin.get_info(name)
            print(f"{info['name']}: {info['describe']} - {info['version']}")
            count = count + 1

        if count == 0:
            print("no plugin")
            sys.exit(0)

        sys.exit(0)

    if args.kcl_plugin_subcmd_name == "init":
        plugin_root = plugin.get_plugin_root()
        if not plugin_root:
            print("# plugin_root: <not found>")
            sys.exit(0)

        names = plugin.get_plugin_names()
        if args.plugin_name in names:
            print(f"{args.plugin_name} exists")
            sys.exit(1)

        plugin.init_plugin(args.plugin_name)
        sys.exit(0)

    if args.kcl_plugin_subcmd_name == "info":
        plugin_root = plugin.get_plugin_root()
        if not plugin_root:
            print("# plugin_root: <not found>")
            sys.exit(0)

        if not args.plugin_name:
            print(f"plugin_root: {plugin.get_plugin_root()}")
            sys.exit(0)

        names = plugin.get_plugin_names()
        if args.plugin_name not in names:
            print(f"{args.plugin_name} not found")
            sys.exit(1)

        info = plugin.get_info(args.plugin_name)

        print(json.dumps(info, indent=4))
        sys.exit(0)

    if args.kcl_plugin_subcmd_name == "gendoc":
        plugin_root = plugin.get_plugin_root()
        if not plugin_root:
            print("# plugin_root: <plugin root not found>")
            sys.exit(0)

        for name in plugin.get_plugin_names():
            plugin.gendoc(name)

        sys.exit(0)

    if args.kcl_plugin_subcmd_name == "version":
        print(plugin.get_plugin_version())
        sys.exit(0)

    # TODO: Using kcl-test
    if args.kcl_plugin_subcmd_name == "test":
        names = plugin.get_plugin_names()
        if args.plugin_name not in names:
            print(f"{args.plugin_name} not found")
            sys.exit(1)

        import pytest

        pytest.main(["-x", plugin.get_plugin_root(args.plugin_name)])
        sys.exit(0)

    parser.print_help()
    plugin_root = plugin.get_plugin_root()
    sys.exit(0)


if __name__ == "__main__":
    Main()



# Copyright 2021 The KCL Authors. All rights reserved.

import os
from pathlib import Path

VERSION = "0.9.0-beta.1"
CHECKSUM = Path(os.path.dirname(__file__)).joinpath("checksum.txt").read_text().strip()



# Copyright 2021 The KCL Authors. All rights reserved.
from copy import deepcopy
from dataclasses import dataclass, field
from typing import Optional, Union, List, Dict
from .decorator import KCLDecoratorObject
from .object import (
    KCLObject,
    KCLObjectType,
    KCLTrueObject,
    KCLFalseObject,
    KCLIntObject,
    KCLFloatObject,
    KCLStringObject,
    KCLBaseTypeObject,
    KCLDictObject,
    KCLSchemaObject,
    KCLNoneObject,
    KCLUndefinedObject,
    KCLTypeKind,
    KCLAnyTypeObject,
    KCLStringTypeObject,
    KCLSchemaReverseFields,
    to_kcl_obj,
)
from .function import KWArg, KCLCompiledFunctionObject

import kclvm.kcl.error as kcl_error
import kclvm.kcl.info as kcl_info
import kclvm.api.object.internal.common as common

from kclvm.internal.util import hash
from kclvm.compiler.check.check_type import (
    type_pack_and_check,
    check_type,
    has_literal_type,
)
from kclvm.kcl.ast import ast

SETTINGS_OUTPUT_KEY = "output_type"
SETTINGS_OUTPUT_STANDALONE = "STANDALONE"
SETTINGS_OUTPUT_INLINE = "INLINE"
SETTINGS_OUTPUT_IGNORE = "IGNORE"

SCHEMA_SETTINGS_ATTR_NAME = "__settings__"
SCHEMA_TYPE_ATTR_NAME = "__schema_type__"
SCHEMA_RUNTIME_TYPE_ATTR_NAME = "__runtime_schema_type__"
MAIN_MODULE_NAME = "__main__"

SCHEMA_SELF_VALUE_KEY = "$schema_self"
SCHEMA_CONFIG_VALUE_KEY = "$schema_config"
SCHEMA_CONFIG_META_KEY = "$schema_config_meta"


class RefGraph:
    """
    Reference graph
    """

    def __init__(self):
        self.adjs = {}

    def _find_node_index(self, node):
        return self.nodeset.index(node)

    def add_node_judge_cycle(self, node, another_node):
        """
        add edge into the schema inheritance graph and check if cyclic inheritance occurs in schema
        """
        if node not in self.adjs:
            self.adjs[node] = []
        if another_node not in self.adjs:
            self.adjs[another_node] = []
        self.adjs[another_node].append(node)
        return self._has_cycle()

    def _has_cycle(self):
        """
        Determine whether the schema inheritance graph is a Directed Acyclic Graph (DAG).
        The detection uses Depth First Search (DFS) algorithm for each node
        in the ergodic graph, and the time complexity is O (V + E),
        V: the total number of detected nodes,
        E: the total number of edges connected by nodes
        """
        visited = {name: 0 for name in self.adjs.keys()}

        def _dfs(name):
            visited[name] = 1
            for adj in self.adjs[name]:
                if visited[adj] == 1:
                    return True
                if visited[adj] == 0:
                    if _dfs(adj):
                        return True
                    else:
                        continue
            visited[name] = 2
            return False

        for name in visited.keys():
            if visited[name] == 0 and _dfs(name):
                return True
        return False


class SchemaTypeRefGraph(RefGraph):
    def get_sub_schemas(self, name: str) -> list:
        """Get all sub schemas by name using BFS"""
        result = []
        if not name:
            return result
        sub_schemas = self.adjs.get(name, [])
        result += sub_schemas
        for sub_schema in sub_schemas:
            result += self.get_sub_schemas(sub_schema)
        return result


class SchemaTypeFactory:
    """
    A schema_type factory used to get schema_type object.
    """

    def __init__(self):
        self._schema_types = {}

    def register(self, name: str, schema_type: "KCLSchemaTypeObject"):
        """
        Register a schema_type with a unique name.

        :param name: Name of the schema_type
        :param schema_type: The schema_type to be registered
        :return: None
        """
        self._schema_types[name] = schema_type

    def get(self, name: str):
        """
        Get and return a schema_type object.

        :param name: Name of the schema_type
        :return: A schema_type object
        """
        schema_type = self._schema_types.get(name)
        if not schema_type:
            raise Exception(f"unknown schema type '{name}'")
        return schema_type


@dataclass
class KCLSchemaIndexSignatureObject(KCLObject):
    key_name: Optional[str] = None
    key_type: str = None
    value_type: str = None
    value: KCLObject = None
    any_other: bool = None
    key_kcl_type: KCLBaseTypeObject = None
    value_kcl_type: KCLBaseTypeObject = None
    node: ast.SchemaIndexSignature = None

    def type(self) -> KCLObjectType:
        """
        Get the object type
        """
        return KCLObjectType.SCHEMA_INDEX_SIGNATURE

    def type_str(self) -> str:
        """
        Get the object type string
        """
        return "IndexSignatureType"

    def def_str(self) -> str:
        return (
            # key_name
            f"[{self.key_name + ': ' if self.key_name else ''}"
            # ...
            + ("..." if self.any_other else "")
            # key_type
            + f"{self.key_type}]: "
            # value_type
            + f"{self.value_type}"
        )


@dataclass
class KCLSchemaAttrObject(KCLObject):
    is_optional: bool = True
    is_final: bool = False
    has_default: bool = False
    attr_type: KCLBaseTypeObject = None
    attr_node: Optional[ast.AST] = None


@dataclass
class KCLSchemaTypeObject(KCLBaseTypeObject):
    name: str = None
    MEMBER_FUNCTIONS = ["instances"]  # Member function list
    name: Optional[str] = None  # Schema name
    __refs__: Optional[list] = field(default_factory=list)  # Instance reference list
    func: Optional[KCLCompiledFunctionObject] = None  # Body functions
    check_fn: Optional[KCLCompiledFunctionObject] = None  # Check function
    is_mixin: bool = False  # Mark is a schema mixin
    is_protocol: bool = False  # Mark is a schema protocol
    is_rule: bool = False  # Mark is a rule block
    is_relaxed: bool = False  # Mark is a relaxed schema
    pkgpath: str = ""  # Schema definition package path
    filename: str = ""  # Definition path location
    doc: str = ""  # Schema definition document string
    runtime_type: Optional[str] = None  # Schema runtime type file_hash + schema_name
    base: Optional["KCLSchemaTypeObject"] = None  # Base schema
    protocol: Optional["KCLSchemaTypeObject"] = None  # Protocol schema
    mixins_names: Optional[List[str]] = field(default_factory=list)
    mixins: List["KCLSchemaTypeObject"] = field(
        default_factory=list
    )  # Schema mixin list
    attrs: Optional[dict] = field(default_factory=dict)  # Schema attributes
    attr_list: Optional[list] = field(default_factory=list)  # Schema attribute order
    attr_obj_map: Dict[Union[str, int, float], Optional[KCLSchemaAttrObject]] = field(
        default_factory=dict
    )  # Schema attribute type map
    settings: Optional[dict] = field(default_factory=dict)  # Schema settings
    decorators: Optional[List[KCLDecoratorObject]] = field(
        default_factory=list
    )  # Schema decorator list
    index_signature: Optional[
        KCLSchemaIndexSignatureObject
    ] = None  # Schema Index signature
    node_ref: Optional[ast.SchemaStmt] = None

    # -----------------
    # Schema eval cache
    # -----------------

    _eval_cache = {}

    def can_add_members(self) -> bool:
        return (
            self.name.endswith("Mixin")
            or self.index_signature is not None
            or self.is_relaxed
        )

    def type(self) -> KCLObjectType:
        """
        Get the object type
        """
        return KCLObjectType.SCHEMA_TYPE

    def type_str(self) -> str:
        """
        Get the object type string
        """
        return (
            self.name
            if (self.pkgpath == "__main__" or not self.pkgpath)
            else f"{self.pkgpath}.{self.name}"
        )

    def type_str_with_pkgpath(self) -> str:
        """
        Get the object type string with pkgpath
        """
        return (
            self.name
            if (self.pkgpath == "__main__" or not self.pkgpath)
            else f"@{self.pkgpath}.{self.name}"
        )

    def type_kind(self):
        return KCLTypeKind.SchemaKind

    @property
    def value(self) -> str:
        """Return the runtime type string"""
        return self.runtime_type

    @property
    def key_type(self) -> str:
        if self.index_signature:
            return self.index_signature.key_kcl_type
        return KCLStringTypeObject()

    @property
    def value_type(self) -> str:
        if self.index_signature:
            return self.index_signature.value_kcl_type
        return KCLAnyTypeObject()

    @property
    def should_add_additional_key(self) -> bool:
        return self.is_relaxed or self.index_signature is not None

    @property
    def file_and_type(self) -> str:
        return self.filename + self.runtime_type

    @property
    def file_and_name(self) -> str:
        return self.filename + self.name

    def is_sub_schema_of(self, base: Union["KCLSchemaTypeObject", str]) -> bool:
        base_type_obj = base
        if not isinstance(base_type_obj, KCLSchemaTypeObject):
            return False
        if (
            self.runtime_type == base.runtime_type
            or self.file_and_name == base.file_and_name
        ):
            return True
        base_ref = self.base
        while base_ref and base_ref.runtime_type != base_type_obj.runtime_type:
            base_ref = base_ref.base
        return True if base_ref else False

    def get_obj_of_attr(self, attr: Union[int, str]) -> Optional[KCLSchemaAttrObject]:
        if attr in self.attr_obj_map:
            return self.attr_obj_map[attr]
        base_ref = self.base
        while base_ref and attr not in base_ref.attr_obj_map:
            base_ref = base_ref.base
        if base_ref:
            return base_ref.attr_obj_map[attr]
        return (
            self.protocol.attr_obj_map[attr]
            if self.protocol and attr in self.protocol.attr_obj_map
            else None
        )

    def get_type_of_attr(self, attr: Union[int, str]) -> Optional[KCLBaseTypeObject]:
        attr_obj = self.get_obj_of_attr(attr)
        return attr_obj.attr_type if attr_obj else None

    def get_node_of_attr(self, attr: Union[int, str]) -> Optional[ast.AST]:
        attr_obj = self.get_obj_of_attr(attr)
        return attr_obj.attr_node if attr_obj else None

    def set_type_of_attr(self, attr: Union[int, str], tpe: KCLBaseTypeObject):
        if attr in self.attr_obj_map:
            self.attr_obj_map[attr].attr_type = tpe
        else:
            self.attr_obj_map[attr] = KCLSchemaAttrObject(attr_type=tpe)

    def set_node_of_attr(self, attr: Union[int, str], ast_node: ast.AST):
        if attr in self.attr_obj_map:
            self.attr_obj_map[attr].attr_node = ast_node
        else:
            self.attr_obj_map[attr] = KCLSchemaAttrObject(attr_node=ast_node)

    def add_decorators(self, decorators: List[KCLDecoratorObject]):
        if not self.decorators:
            self.decorators = []
        self.decorators += decorators

    def set_func(self, func: KCLCompiledFunctionObject):
        assert isinstance(func, KCLCompiledFunctionObject)
        self.func = func

    def set_check_func(self, check_fn: Optional[KCLCompiledFunctionObject]):
        if check_fn:
            assert isinstance(check_fn, KCLCompiledFunctionObject)
            self.check_fn = check_fn
        else:
            self.check_fn = None

    @staticmethod
    def schema_runtime_type(tpe: str, filename=""):
        return f"runtime_type_{hash(filename)}_{tpe}"

    @staticmethod
    def new(
        name: str,
        base: Optional["KCLSchemaTypeObject"] = None,
        protocol: Optional["KCLSchemaTypeObject"] = None,
        filename: str = "",
        is_relaxed: bool = False,
        is_mixin: bool = False,
        pkgpath: str = "",
        attr_list: list = None,
        index_signature: Optional[KCLSchemaIndexSignatureObject] = None,
        vm=None,
    ):
        if not name:
            kcl_error.report_exception(
                err_type=kcl_error.ErrType.EvaluationError_TYPE,
                arg_msg="schema name can't be None",
            )
        runtime_type = KCLSchemaTypeObject.schema_runtime_type(name, pkgpath)
        settings = {
            SETTINGS_OUTPUT_KEY: SETTINGS_OUTPUT_INLINE
            if not kcl_info.isprivate_field(name)
            else SETTINGS_OUTPUT_IGNORE,
            KCLSchemaReverseFields.NAME: name,
            KCLSchemaReverseFields.TYPE: f"{pkgpath}.{name}",
            KCLSchemaReverseFields.PKG_PATH: pkgpath,
        }
        obj = KCLSchemaTypeObject(
            pkgpath=pkgpath,
            name=name,
            base=base,
            protocol=protocol,
            settings=settings,
            filename=filename,
            is_relaxed=is_relaxed,
            is_mixin=is_mixin,
            runtime_type=runtime_type,
            attrs={attr: KCLUndefinedObject.instance() for attr in attr_list},
            attr_list=attr_list,
            index_signature=index_signature,
        )
        return obj

    def update_mixins(self, vm=None):
        """Get mixins by name"""
        if not self.mixins and self.mixins_names:
            self.mixins = []
            for mixin_name in self.mixins_names:
                if "." in mixin_name:  # pkg.Schema
                    schema_type_obj = vm.find_schema_type(mixin_name)
                    if schema_type_obj and isinstance(
                        schema_type_obj, KCLSchemaTypeObject
                    ):
                        self.mixins.append(schema_type_obj)
                    else:
                        kcl_error.report_exception(
                            err_type=kcl_error.ErrType.EvaluationError_TYPE,
                            arg_msg="name '{}' is not defined".format(mixin_name),
                        )
                else:
                    if mixin_name not in vm.frames[0].globals:
                        kcl_error.report_exception(
                            err_type=kcl_error.ErrType.EvaluationError_TYPE,
                            arg_msg="name '{}' is not defined".format(mixin_name),
                        )
                    schema_type_obj = vm.frames[0].globals[mixin_name]
                    self.mixins.append(schema_type_obj)

    def has_base(self):
        return self.base and isinstance(self.base, KCLSchemaTypeObject)

    def new_empty_instance(self) -> KCLSchemaObject:
        return KCLSchemaObject(
            attrs={SCHEMA_SETTINGS_ATTR_NAME: to_kcl_obj(self.settings)},
            pkgpath=self.pkgpath,
            name=self.name,
            runtime_type=self.runtime_type,
        )

    def new_instance(
        self,
        config: Union[dict, KCLDictObject],
        config_meta: dict,
        args: List[KCLObject],
        kwargs: List[KWArg],
        vm,
        finalize: bool = False,
    ):
        from kclvm.vm.runtime.evaluator import SchemaEvalContext

        context = SchemaEvalContext.build_from_vm(
            vm=vm,
            type_obj=self,
            schema_obj=self.new_empty_instance(),
            config=config,
            config_meta=config_meta,
            args=args,
            kwargs=kwargs,
        )
        # Save origin eval context
        org_eval_ctx = vm.lazy_eval_ctx
        vm.lazy_eval_ctx = context
        # Reset the eval status before the evaluation
        context.eval_reset()
        # Get all schema attribute value place holders using the schema type object and the cache
        if self.file_and_type in self._eval_cache:
            context.place_holder_map = self._eval_cache[self.file_and_type]
        else:
            context.get_all_place_holders()
            self._eval_cache[self.file_and_type] = context.place_holder_map
        # New a schema instance using the type object
        context.schema_obj = self._new_instance(
            config,
            config_meta,
            args,
            kwargs,
            vm,
            # Put the schema instance reference
            inst=context.schema_obj,
            finalize=finalize,
        )
        # Reset the eval status after the evaluation
        context.eval_reset()
        # Reload origin eval context
        vm.lazy_eval_ctx = org_eval_ctx
        # Return the schema instance
        context.schema_obj.config_keys = (
            set(config.value.keys())
            if isinstance(config, KCLDictObject)
            else set(config.keys())
        )
        return context.schema_obj

    def _new_instance(
        self,
        config: Union[dict, KCLDictObject],
        config_meta: dict,
        args: List[KCLObject],
        kwargs: List[KWArg],
        vm,
        inst: KCLSchemaObject = None,
        is_sub_schema: bool = False,
        finalize: bool = False,
    ) -> KCLSchemaObject:
        self.do_args_type_check(args, kwargs, config_meta, vm)
        inst = inst or self.new_empty_instance()
        inst.instance_pkgpath = vm.ctx.pkgpath
        for decorator in self.decorators:
            inst.add_decorator(self.name, decorator=decorator)
        if self.base and isinstance(self.base, KCLSchemaTypeObject):
            inst = self.base._new_instance(
                config,
                config_meta,
                [],
                [],
                vm,
                inst,
                is_sub_schema=True,
                finalize=finalize,
            )
        # Record all schema attributes
        inst.union_with(self.attrs, should_idempotent_check=False)
        for mixin in self.mixins:
            inst.union_with(mixin.attrs, should_idempotent_check=False)
        # Record the schema name, runtime_type and relaxed
        inst.update_info(self.name, self.runtime_type, self.is_relaxed)
        vm.push_frame_using_callable(
            self.pkgpath,
            self.func,
            (args if args else []) + [config_meta, config, inst],
            kwargs,
            args_len=len(args),
        )
        # Run the schema compiled function body
        vm.run(run_current=True, ignore_nop=True)

        if SCHEMA_SETTINGS_ATTR_NAME not in inst:
            inst.update({SCHEMA_SETTINGS_ATTR_NAME: self.settings})
        if SCHEMA_SETTINGS_ATTR_NAME in config:
            inst.update(
                {SCHEMA_SETTINGS_ATTR_NAME: config.get(SCHEMA_SETTINGS_ATTR_NAME)}
            )

        # Add settings attr
        if not self.attrs:
            self.attrs = {}
        self.attrs[SCHEMA_SETTINGS_ATTR_NAME] = (
            inst.get(SCHEMA_SETTINGS_ATTR_NAME) or self.settings
        )

        # Do relaxed schema check and config patch
        relaxed_keys = self.do_relaxed_check(
            inst, config_meta, config, is_sub_schema, vm
        )

        # Record the schema name, runtime_type and relaxed
        inst.update_info(self.name, self.runtime_type, self.is_relaxed)

        self.update_mixins(vm=vm)

        # Do all mixins expand execution after schema context
        if self.mixins:
            for mixin in self.mixins:
                inst = mixin._new_instance(
                    config,
                    config_meta,
                    [],
                    [],
                    vm,
                    inst,
                    is_sub_schema=True,
                    finalize=finalize,
                )

        # Record the schema name, runtime_type and relaxed
        inst.update_info(self.name, self.runtime_type, self.is_relaxed)

        # Record schema instance
        if not self.__refs__:
            self.__refs__ = []
        self.__refs__.append(inst)

        # Deal schema stmt queue
        if not is_sub_schema and inst.stmt_buffer():
            buffers = inst.stmt_buffer()
            func = KCLCompiledFunctionObject(
                name=self.func.name,
                params=self.func.params,
                names=self.func.names,
                constants=self.func.constants,
            )
            relaxed = inst.is_relaxed
            for buffer in buffers:
                is_relaxed, pkg_path, codes = buffer
                func.instructions = codes
                inst.is_relaxed = is_relaxed
                vm.push_frame_using_callable(
                    pkg_path,
                    func,
                    (args if args else []) + [config_meta, config, inst],
                    kwargs,
                    args_len=len(args),
                )
                # Run the schema compiled function body
                vm.run(run_current=True, ignore_nop=True)
            inst.is_relaxed = relaxed
        # Run all decorators
        inst.run_all_decorators()
        # Do all checks
        if not is_sub_schema:
            inst.operation_map = {
                KCLSchemaReverseFields.SETTINGS: ast.ConfigEntryOperation.OVERRIDE
            }
            inst.update_attr_op_using_obj(config)
            if finalize:
                inst.check_optional_attrs(True)
            self.do_check(inst, config_meta, config, relaxed_keys, vm)
        # Return the schema object
        return inst

    def do_args_type_check(
        self,
        args: List[KCLObject],
        kwargs: List[KWArg],
        config_meta: dict,
        vm=None,
    ):
        """Check args type"""

        def check_arg_type(arg_name: str, value: KCLObject, expected_type: str):
            checked, value_tpe = check_type(value, expected_type, vm)
            if not checked:
                if has_literal_type([expected_type]):
                    if isinstance(
                        value,
                        (
                            KCLNoneObject,
                            KCLTrueObject,
                            KCLFalseObject,
                            KCLIntObject,
                            KCLFloatObject,
                        ),
                    ):
                        value_tpe = f"{value_tpe}({value.value})"
                    elif isinstance(value, KCLStringObject):
                        value_tpe = f'{value_tpe}("{value.value}")'

                conf_filename, conf_line, conf_column = (
                    config_meta.get("$filename"),
                    config_meta.get("$lineno"),
                    config_meta.get("$columnno"),
                )
                kcl_error.report_exception(
                    err_type=kcl_error.ErrType.TypeError_Runtime_TYPE,
                    file_msgs=[
                        kcl_error.ErrFileMsg(
                            filename=conf_filename,
                            line_no=conf_line,
                            col_no=conf_column,
                        )
                    ],
                    arg_msg='argument "{}" expect {}, got {}'.format(
                        arg_name,
                        common.get_tpes_str([expected_type]).replace("@", ""),
                        common.get_class_name(value_tpe),
                    ),
                )

        if self.func.params:
            for i, value in enumerate(args or []):
                arg_name = self.func.params[i].name
                expected_type = self.func.params[i].type_annotation
                check_arg_type(arg_name, value, expected_type)

            for kwarg in kwargs or []:
                arg_name = kwarg.name.value
                value = kwarg.value
                if arg_name not in [p.name for p in self.func.params]:
                    kcl_error.report_exception(
                        err_type=kcl_error.ErrType.EvaluationError_TYPE,
                        arg_msg=f"schema arguments got an unexpected keyword argument '{arg_name}'",
                    )
                expected_types = [
                    p.type_annotation for p in self.func.params if arg_name == p.name
                ]
                expected_type = expected_types[0] if expected_types else None
                check_arg_type(arg_name, value, expected_type)

    def do_relaxed_check(
        self,
        inst: KCLSchemaObject,
        config_meta: dict,
        config: Union[dict, KCLDictObject],
        is_sub_schema: bool,
        vm,
    ) -> List[str]:
        """Do relaxed schema check and config patch"""
        relaxed_keys = []
        if not is_sub_schema:
            config_native = (
                config.value if isinstance(config, KCLDictObject) else config
            )
            config_meta_native = config_meta
            relaxed_keys = [
                key for key in config_native if key not in inst.value.keys()
            ]
            if self.protocol:
                relaxed_keys = [
                    key for key in relaxed_keys if key not in self.protocol.attr_list
                ]
            if self.is_relaxed or self.index_signature:
                filename = vm.get_filename()
                if self.index_signature and not self.index_signature.any_other:
                    for key in inst.value.keys():
                        if key != SCHEMA_SETTINGS_ATTR_NAME:
                            value = inst.get(key)
                            checked, _ = check_type(
                                value,
                                self.index_signature.value_type,
                                vm=vm,
                            )
                            if not checked:
                                kcl_error.report_exception(
                                    err_type=kcl_error.ErrType.EvaluationError_TYPE,
                                    file_msgs=[kcl_error.ErrFileMsg(filename=filename)],
                                    arg_msg=f"the type '{value.type_str()}' of schema attribute '{key}' "
                                    f"does not meet the index signature definition {self.index_signature.def_str()}",
                                )
                for key in relaxed_keys:
                    lineno, columnno = None, None
                    if key in config_meta_native:
                        lineno, columnno = (
                            config_meta_native[key].get("lineno"),
                            config_meta_native[key].get("columnno"),
                        )
                    value = config.get(key)
                    if self.index_signature and self.index_signature.value_type:
                        types = [self.index_signature.value_type]
                        from kclvm.vm.runtime.evaluator import union

                        value = type_pack_and_check(
                            union(
                                deepcopy(self.index_signature.value),
                                value,
                                should_idempotent_check=True,
                                vm=vm,
                            ),
                            types,
                            filename,
                            lineno,
                            columnno,
                            vm=vm,
                            config_meta=config_meta,
                        )
                    inst.update({key: value})

            elif relaxed_keys:
                lineno, columnno = None, None
                if relaxed_keys[0] in config_meta_native:
                    lineno, columnno = (
                        config_meta_native[relaxed_keys[0]].get("lineno"),
                        config_meta_native[relaxed_keys[0]].get("columnno"),
                    )
                kcl_error.report_exception(
                    err_type=kcl_error.ErrType.CannotAddMembers_Runtime_TYPE,
                    file_msgs=[
                        kcl_error.ErrFileMsg(
                            filename=vm.get_filename(), line_no=lineno, col_no=columnno
                        )
                    ],
                    arg_msg=kcl_error.CANNOT_ADD_MEMBERS_MSG.format(
                        ",".join([str(k) for k in relaxed_keys]), self.name
                    ),
                )
        return relaxed_keys

    def do_check(
        self,
        inst: KCLSchemaObject,
        config_meta: dict,
        config: Union[dict, KCLDictObject],
        relaxed_keys: List[str],
        vm,
    ):
        assert inst, f"{inst}"
        assert vm

        def call_check_fn(local_name: str = None, local_value: KCLObject = None):
            if self.check_fn:
                vm.push_frame_using_callable(
                    self.pkgpath, self.check_fn, [config_meta, config, inst], []
                )
                if local_name and local_value:
                    vm.update_local(local_name, local_value)
                vm.run(run_current=True, ignore_nop=True)

        # check base
        if self.base and isinstance(self.base, KCLSchemaTypeObject):
            self.base.do_check(inst, config_meta, config, relaxed_keys, vm)

        # check mixin
        for mixin in self.mixins or []:
            mixin.do_check(inst, config_meta, config, relaxed_keys, vm)

        # check self
        if self.index_signature and self.index_signature.key_name and relaxed_keys:
            # For loop index signature attributes
            for key in relaxed_keys:
                call_check_fn(self.index_signature.key_name, to_kcl_obj(key))
        else:
            call_check_fn()

    # Member Functions

    def instances(self, main_pkg: bool = True):
        """Get all schema instances of self type and sub types"""
        if not self.__refs__:
            self.__refs__ = []
        return deepcopy(
            [
                inst
                for inst in self.__refs__
                if inst.instance_pkgpath == MAIN_MODULE_NAME
            ]
            if main_pkg
            else self.__refs__
        )

    def get_member_method(self, name: str):
        from .function import KCLMemberFunctionObject

        if not name:
            kcl_error.report_exception(
                err_type=kcl_error.ErrType.EvaluationError_TYPE,
                arg_msg="kcl string object member name can't be empty or None",
            )
        if name not in self.MEMBER_FUNCTIONS:
            kcl_error.report_exception(
                err_type=kcl_error.ErrType.AttributeError_TYPE,
                arg_msg=f"attribute '{name}' not found",
            )
        return KCLMemberFunctionObject(obj=self, name=name)

    def call_member_method(self, name: str, *args, **kwargs):
        if not hasattr(self, name) and name not in self.MEMBER_FUNCTIONS:
            kcl_error.report_exception(
                err_type=kcl_error.ErrType.AttributeError_TYPE,
                arg_msg=f"attribute '{name}' not found",
            )
        return getattr(self, name).__call__(*args, **kwargs)


@dataclass
class KCLSchemaDefTypeObject(KCLBaseTypeObject):
    """Schema definition type denotes the schema definition type used in normal expressions.

    - `Person` of `data = Person.instances()` is a schema def type.
    - `person` of `person = Person {}` is a schema type.
    """

    schema_type: KCLSchemaTypeObject

    def type_str(self) -> str:
        """Get the object type"""
        return self.schema_type.type_str() if self.schema_type else super().type_str()

    def type_kind(self) -> int:
        """Get the"""
        return KCLTypeKind.SchemaDefKind



import pathlib
from typing import Tuple, Union, Optional, List, Any


class Message:
    """This class represent a message to be issued by the reporters"""

    def __init__(
        self,
        msg_id: str,
        file: Optional[Union[str, pathlib.PosixPath]],
        msg: str,
        source_code: str,
        pos: Tuple[int, int],
        arguments: List[Any],
    ):
        self.msg_id = msg_id
        self.level = msg_id[0]
        self.file = str(file) if file else None
        self.msg = msg
        self.source_code = source_code
        self.pos = pos
        self.arguments = [str(arg) for arg in arguments]

    def __str__(self):
        return (
            f"{self.file}:{self.pos[0]}:{self.pos[1]}: {self.msg_id} {self.msg}\n"
            + self.source_code
            + "\n"
            + (self.pos[1] - 1) * " "
            + "^"
        )

    def __eq__(self, other):
        return (
            self.msg_id,
            self.file,
            self.msg,
            self.source_code,
            self.pos,
            self.arguments,
        ) == (
            other.msg_id,
            other.file,
            other.msg,
            other.source_code,
            other.pos,
            other.arguments,
        )



from ruamel.yaml import YAML
import io

from abc import ABC, abstractmethod

import kclvm.tools.docs.pb as pb
import kclvm.tools.docs.model_pb2 as model
import kclvm.tools.docs.factory as factory
import kclvm.tools.docs.formats as doc_formats
import kclvm.tools.docs.templater as templater


# ---------------------------------------------------
# Constants
# ---------------------------------------------------


ENDLINE = "\n"


# -------------------------------------------------------
# Document writer from KCL ModuleDoc type ot markdown type
# -------------------------------------------------------


class ModuleDocWriter(ABC):
    """Module document writer class.

    This class provides basic functions for write KCL Module to the IO.
    """

    def __init__(self, out: io.TextIOBase):
        self.out: io.TextIOBase = out

    @abstractmethod
    def write_doc(self, module: model.ModuleDoc) -> None:
        pass

    def write(self, content: str = ""):
        """Write the content to the string io"""
        self.out.write(content)

    def writeln(self, content: str = ""):
        """Write the content with an endline to the string io"""
        self.write(content + ENDLINE)


class JsonDocWriter(ModuleDocWriter):
    """JSON document writer based on ModuleDocWriter.

    The class is used to generate JSON file from KCL module document string.
    """

    def _dump_json_str(self, module: model.ModuleDoc) -> str:
        return pb.ToJson(module)

    def write_doc(self, module: model.ModuleDoc) -> None:
        json_str = self._dump_json_str(module)
        self.write(json_str)


class YamlDocWriter(JsonDocWriter):
    """YAML document writer based on JsonDocWriter.

    The class is used to generate YAML file from KCL module document string.
    """

    def write_doc(self, module: model.ModuleDoc) -> None:
        def set_style(d, flow):
            if isinstance(d, dict):
                if flow:
                    d.fa.set_flow_style()
                else:
                    d.fa.set_block_style()
                for k in d:
                    set_style(d[k], flow)
            elif isinstance(d, list):
                if flow:
                    d.fa.set_flow_style()
                else:
                    d.fa.set_block_style()
                for item in d:
                    set_style(item, flow)

        json_str = self._dump_json_str(module)
        yaml = YAML()
        data = yaml.load(json_str)
        set_style(data, flow=False)
        yaml.dump(data, self.out)


class MarkDownDocWriter(ModuleDocWriter):
    """Markdown document writer based on KCLModuleDocWriter.

    The class is used to generate markdown file from KCL module document string.
    """

    def write_doc(self, module: model.ModuleDoc) -> None:
        """
        Write document
        """
        self.writeln(templater.md_module_doc_templater(module))


factory = factory.IOFactory()
factory.register_format(doc_formats.KCLDocFormat.JSON, JsonDocWriter)
factory.register_format(doc_formats.KCLDocFormat.YAML, YamlDocWriter)
factory.register_format(doc_formats.KCLDocFormat.MARKDOWN, MarkDownDocWriter)



# Copyright 2021 The KCL Authors. All rights reserved.

from .object import *
from .function import *
from .schema import *
from .decorator import *
from .bytecode import *

__all__ = [
    "KCLObject",
    "KCLBaseTypeObject",
    "KCLLiteralObject",
    "KCLIntObject",
    "KCLFloatObject",
    "KCLStringObject",
    "KCLNameConstantObject",
    "KCLTrueObject",
    "KCLFalseObject",
    "KCLNoneObject",
    "KCLBuiltinTypeObject",
    "KCLListTypeObject",
    "KCLDictTypeObject",
    "KCLUnionTypeObject",
    "KCLSchemaTypeObject",
    "KCLNumberMultiplierTypeObject",
    "KCLSchemaObject",
    "KCLSchemaIndexSignatureObject",
    "KCLDictObject",
    "KCLListObject",
    "KCLClosureObject",
    "KCLFunctionObject",
    "KCLBuiltinFunctionObject",
    "KCLMemberFunctionObject",
    "KCLCompiledFunctionObject",
    "KCLUnpackObject",
    "KCLSliceObject",
    "KCLObjectType",
    "KCLIterObject",
    "KCLModuleObject",
    "KCLDecoratorObject",
    "KWArg",
    "SCHEMA_SELF_VALUE_KEY",
    "SCHEMA_CONFIG_VALUE_KEY",
    "Parameter",
    "KCLFunctionTypeObject",
    "to_kcl_obj",
    "to_python_obj",
    "KCLBytecode",
    "KCLProgram",
    "KCLResult",
    "KCLTypeKind",
    "SchemaTypeRefGraph",
]



# Copyright 2021 The KCL Authors. All rights reserved.

from typing import List, Union, cast

import kclvm.kcl.error as kcl_error
import kclvm.kcl.ast as ast
from kclvm.api.object import (
    KCLObject,
    KCLSchemaObject,
    KCLSchemaTypeObject,
    KCLListObject,
    KCLDictObject,
    KCLNoneObject,
    KCLUndefinedObject,
    KCLStringObject,
    KCLLiteralObject,
    KCLSchemaConfigObject,
    KCLConfigObjectMixin,
    to_python_obj,
    to_kcl_obj,
)
from kclvm.api.object.internal import Undefined, UndefinedType
from kclvm.compiler.check.check_type import check_type_builtin
from kclvm.unification import value_subsume


def override_config_attr(
    attr: str,
    obj: Union[KCLDictObject, KCLSchemaObject],
    config_value: KCLObject,
    index: int = None,
):
    if index is None or index < 0:
        obj.update_key_value(attr, config_value)
    else:
        obj.list_key_override(attr, config_value, index)


def insert_config_attr(
    attr: str,
    obj: Union[KCLDictObject, KCLSchemaObject],
    config_value: KCLObject,
    index: int = None,
):
    if index is not None and index >= 0:
        config_value = KCLListObject(items=[config_value])
    obj.insert_with_key(attr, config_value, index)


def resolve_schema_obj(
    schema_obj: KCLSchemaObject, keys: set, vm=None
) -> KCLSchemaObject:
    """Using a schema object config to resolve and generate a new schema"""
    if not vm or not schema_obj or not isinstance(schema_obj, KCLSchemaObject):
        return schema_obj
    schema_type_obj = cast(
        KCLSchemaTypeObject,
        vm.all_schema_types.get(f"{schema_obj.pkgpath}.{schema_obj.name}"),
    )
    if not schema_type_obj:
        return schema_obj
    filename, line, column = vm.get_info()
    config_meta = {
        "$filename": filename,
        "$lineno": line,
        "$columnno": column,
    }
    config = KCLSchemaConfigObject(
        value={k: schema_obj.attrs[k] for k in schema_obj.attrs if k in keys},
        operation_map=schema_obj.operation_map,
    )
    return schema_type_obj.new_instance(config, config_meta, [], [], vm)


def do_union(
    obj: KCLObject,
    delta: KCLObject,
    should_list_override: bool = False,
    should_idempotent_check: bool = False,
    should_config_resolve: bool = False,
    vm=None,
) -> KCLObject:
    """
    Union delta to obj recursively
    """
    obj_tpe = obj.type_str()
    delta_tpe = delta.type_str()
    if isinstance(delta, KCLStringObject):
        delta_tpe = "str"
    if isinstance(obj, KCLListObject):
        if not isinstance(delta, KCLListObject):
            kcl_error.report_exception(
                err_type=kcl_error.ErrType.TypeError_Runtime_TYPE,
                arg_msg=f"union failure, expect list, got {delta_tpe}",
            )

        length = (
            len(obj.value) if len(obj.value) > len(delta.value) else len(delta.value)
        )
        if should_list_override:
            return delta
        result_list = obj
        for idx in range(length):
            if idx >= len(obj.value):
                result_list.value.append(delta.value[idx])
            elif idx < len(delta.value):
                result_list.value[idx] = union(
                    result_list.value[idx],
                    delta.value[idx],
                    should_list_override=should_list_override,
                    should_idempotent_check=should_idempotent_check,
                    should_config_resolve=should_config_resolve,
                    vm=vm,
                )
        return result_list
    if isinstance(obj, KCLDictObject):
        if not isinstance(delta, KCLDictObject):
            kcl_error.report_exception(
                err_type=kcl_error.ErrType.TypeError_Runtime_TYPE,
                arg_msg=f"union failure, expect dict, got {delta_tpe}",
            )
        result_dict = obj
        if isinstance(obj, KCLConfigObjectMixin):
            obj.update_attr_op_using_obj(delta)
        for k in delta.value:
            operation = (
                delta.get_operation(k)
                if isinstance(delta, KCLConfigObjectMixin)
                else ast.ConfigEntryOperation.UNION
            )
            insert_index = (
                delta.get_insert_index(k)
                if isinstance(delta, KCLConfigObjectMixin)
                else None
            )
            if k not in obj.value:
                result_dict.value[k] = delta.value[k]
            else:
                if operation == ast.ConfigEntryOperation.OVERRIDE:
                    override_config_attr(k, result_dict, delta.value[k], insert_index)
                if operation == ast.ConfigEntryOperation.INSERT:
                    insert_config_attr(k, result_dict, delta.value[k], insert_index)
                else:
                    if (
                        should_idempotent_check
                        and k in obj.value
                        and not value_subsume(delta.value[k], obj.value[k], False)
                    ):
                        kcl_error.report_exception(
                            err_type=kcl_error.ErrType.EvaluationError_TYPE,
                            arg_msg=f"conflicting values on the attribute '{k}' between "
                            f"{to_python_obj(delta)} and {to_python_obj(obj)}",
                        )
                    result_dict.value[k] = union(
                        obj.value[k],
                        delta.value[k],
                        should_list_override=should_list_override,
                        should_idempotent_check=should_idempotent_check,
                        should_config_resolve=should_config_resolve,
                        vm=vm,
                    )
        return result_dict
    if isinstance(obj, KCLSchemaObject):
        delta_dict = {}
        if isinstance(delta, (KCLDictObject, KCLSchemaObject)):
            delta_dict = delta.value
        else:
            kcl_error.report_exception(
                err_type=kcl_error.ErrType.TypeError_Runtime_TYPE,
                arg_msg=f"union failure, expect {obj_tpe}, got {delta_tpe}",
            )
        if should_config_resolve:
            common_keys = obj.config_keys | delta.config_keys
        if isinstance(obj, KCLConfigObjectMixin):
            obj.update_attr_op_using_obj(delta)
        for k in delta_dict:
            if should_config_resolve and k not in common_keys:
                continue
            operation = (
                delta.get_operation(k)
                if isinstance(delta, KCLConfigObjectMixin)
                else ast.ConfigEntryOperation.UNION
            )
            insert_index = (
                delta.get_insert_index(k)
                if isinstance(delta, KCLConfigObjectMixin)
                else None
            )
            if (
                should_config_resolve
                and k not in obj.attrs
                and not obj.should_add_attr(k)
            ):
                kcl_error.report_exception(
                    err_type=kcl_error.ErrType.CannotAddMembers_Runtime_TYPE,
                    arg_msg=kcl_error.CANNOT_ADD_MEMBERS_MSG.format(k, obj.name),
                )
            if operation == ast.ConfigEntryOperation.OVERRIDE:
                override_config_attr(k, obj, delta_dict[k], insert_index)
            if operation == ast.ConfigEntryOperation.INSERT:
                insert_config_attr(k, obj, delta_dict[k], insert_index)
            else:
                if (
                    should_idempotent_check
                    and k in obj.attrs
                    and not value_subsume(delta_dict[k], obj.attrs[k], False)
                ):
                    kcl_error.report_exception(
                        err_type=kcl_error.ErrType.EvaluationError_TYPE,
                        arg_msg=f"conflicting values on the attribute '{k}' between "
                        f"{to_python_obj(obj)} and {to_python_obj(delta_dict)}",
                    )
                obj.attrs[k] = union(
                    obj.attrs.get(k),
                    delta_dict[k],
                    should_list_override=should_list_override,
                    should_idempotent_check=should_idempotent_check,
                    should_config_resolve=should_config_resolve,
                    vm=vm,
                )
            # Do type check and pack
            if isinstance(obj.attrs[k], KCLLiteralObject):
                check_type_builtin(obj.attrs[k], obj.get_attr_type(k))
        if should_config_resolve:
            obj = resolve_schema_obj(obj, common_keys, vm=vm)
            obj.config_keys = common_keys
        return obj
    if type(obj) != type(delta):
        kcl_error.report_exception(
            err_type=kcl_error.ErrType.TypeError_Runtime_TYPE,
            arg_msg=f"union failure, expect {obj_tpe}, got {delta_tpe}",
        )
    return delta


def union(
    obj: KCLObject,
    delta: KCLObject,
    or_mode: bool = False,
    should_list_override: bool = False,
    should_idempotent_check: bool = False,
    should_config_resolve: bool = False,
    vm=None,
) -> KCLObject:
    if (
        obj is None
        or obj is Undefined
        or isinstance(obj, UndefinedType)
        or isinstance(obj, (KCLNoneObject, KCLUndefinedObject))
    ):
        return delta
    if (
        delta is None
        or delta is Undefined
        or isinstance(delta, UndefinedType)
        or isinstance(delta, (KCLNoneObject, KCLUndefinedObject))
    ):
        return obj
    if isinstance(obj, (KCLListObject, KCLSchemaObject, KCLDictObject)) or isinstance(
        delta, (KCLListObject, KCLSchemaObject, KCLDictObject)
    ):
        return do_union(
            obj,
            delta,
            should_list_override=should_list_override,
            should_idempotent_check=should_idempotent_check,
            should_config_resolve=should_config_resolve,
            vm=vm,
        )
    if or_mode:
        return to_kcl_obj(to_python_obj(obj) | to_python_obj(delta))
    else:
        return obj if isinstance(delta, (KCLNoneObject, KCLUndefinedObject)) else delta


def merge(objs: List[KCLObject], vm=None) -> KCLObject:
    """Merge all objects recursively

    - literal variables, override
    - list variables, override
    - dict/schema variables, union
    """
    initial_object = KCLNoneObject.instance()
    if not objs:
        return initial_object
    for obj in objs:
        if not obj or isinstance(obj, (KCLNoneObject, KCLUndefinedObject)):
            continue
        initial_object = union(initial_object, obj, should_list_override=True, vm=vm)
    return initial_object



import hashlib
import typing


class dotdict(dict):
    """dot.notation access to dictionary attributes"""

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


def hash(input_str):
    return hashlib.md5(input_str.encode("utf-8")).hexdigest()


def merge_option_same_keys(args):
    """
    merge kcl -D and -Y argument with the same keys
    """
    if not args:
        return {}
    return {k: v for k, v in args}


def safe_call(fn: typing.Callable, *args, **kwargs) -> (typing.Any, Exception):
    result = None
    try:
        result = fn(*args, **kwargs)
        return result, None
    except Exception as err:
        return result, err



from .override import (
    ApplyOverrides,
    FixModuleOverride,
    PrintOverridesAST,
    OverrideInfo,
    override_file,
)


__all__ = [
    "ApplyOverrides",
    "FixModuleOverride",
    "PrintOverridesAST",
    "OverrideInfo",
    "override_file",
]



# Copyright 2021 The KCL Authors. All rights reserved.


class UndefinedType:
    def __repr__(self):
        return self.__str__()

    def __str__(self):
        """Built-in str(Undefined) like str(None)"""
        return "Undefined"

    def __bool__(self):
        """Built-in bool(Undefined) like bool(None)"""
        return False

    @staticmethod
    def type_str():
        """Error message show type"""
        return "UndefinedType"

    @property
    def value(self):
        return self._value

    def __init__(self):
        self._value = None


Undefined = UndefinedType()



# Copyright 2021 The KCL Authors. All rights reserved.

import io
from dataclasses import dataclass
from typing import List

import kclvm.kcl.ast as ast
import kclvm.compiler.parser as parser

from .printer import PrintAST


NEWLINE = "\n"
MOCK_SCHEMA = "MockSchema"


@dataclass
class SchemaRuleCodeSnippet:
    """Schema and rule code snippet structure"""

    schema: str = ""
    rule: str = ""


def splice_schema_with_rule(code_snippets: List[SchemaRuleCodeSnippet]) -> str:
    """Splice schema with rule code

    Returns a string result denoting the splicing code.

    Parameters
    ----------
    code_snippets : List[SchemaRuleCodeSnippet]
        A list of schema and rule code snippet structure
    """
    if not isinstance(code_snippets, list):
        raise ValueError(f"Invalid parameter {code_snippets}, expected list")
    with io.StringIO() as buf:
        for code in code_snippets:
            if not code or not isinstance(code, SchemaRuleCodeSnippet):
                raise ValueError(
                    f"Invalid parameter {code}, expected SchemaRuleCodeSnippet"
                )
            module = parser.ParseFile(
                "<schema>.k", code.schema, mode=parser.ParseMode.ParseComments
            )
            module_schema_rule = parser.ParseFile(
                "<rule>.k",
                build_rule_check_block_str(
                    MOCK_SCHEMA,
                    code.rule,
                ),
                mode=parser.ParseMode.ParseComments,
            )
            schema_rule_list = module_schema_rule.GetSchemaList()
            if schema_rule_list:
                for i, stmt in enumerate(module.body):
                    if isinstance(module.body[i], ast.SchemaStmt):
                        module.body[i].checks = schema_rule_list[0].checks
                        for comment in module_schema_rule.comments:
                            comment.line = module.body[i].end_line + 1
                            module.comments.append(comment)
            PrintAST(module, buf)
        return buf.getvalue().rstrip(NEWLINE) + NEWLINE


def add_indent_to_code_string(code: str, indent: int = 4) -> str:
    """Add indent to code string"""
    if not code or not isinstance(code, str):
        return ""
    lines = code.split(NEWLINE)
    return NEWLINE.join([" " * indent + line for line in lines])


def build_rule_check_block_str(schema_name: str, rule_code: str) -> str:
    """Build rule check block string using the rule code string"""
    if not schema_name or not isinstance(rule_code, str):
        return ""
    if not rule_code or not isinstance(rule_code, str):
        return ""
    return (
        f"schema {schema_name}:"
        + NEWLINE
        + add_indent_to_code_string("check:", 4)
        + NEWLINE
        + add_indent_to_code_string(rule_code, 8)
    )



# Copyright 2021 The KCL Authors. All rights reserved.

import io

from typing import List, Dict, Set

import kclvm.compiler.parser as parser
import kclvm.api.object as objpkg
import kclvm.kcl.ast as ast
import kclvm.kcl.types as types
import kcl_lib.api.spec_pb2 as pb2
import kclvm.tools.printer as printer


def get_schema_type_from_code(
    file: str, code: str, schema_name: str = None
) -> List[pb2.KclType]:
    """Get schema types from a kcl file or code.

    Parameters
    ----------
    file: str
        The kcl filename
    code: str
        The kcl code string
    schema_name: str
        The schema name got, when the schema name is empty, all schemas are returned.

    Returns
    -------
    schema_types: List[pb2.KclType]

    KclType:
        string type = 1;                     // schema, dict, list, str, int, float, bool, int(), float() str(), bool()
        repeated KclType union_types = 2 ;   // union types
        string default = 3;                  // default value

        string schema_name = 4;              // schema name
        string schema_doc = 5;               // schema doc
        map<string, KclType> properties = 6; // schema properties
        repeated string required = 7;        // required schema properties, [property_name1, property_name2]

        KclType key = 8;                     // dict key type
        KclType item = 9;                    // dict/list item type

        int32 line = 10;
    """
    result = []
    program = parser.LoadProgram(
        file or "<input>", k_code_list=[code] if code else None
    )
    for name, o in types.ResolveProgram(program).main_scope.elems.items():
        if isinstance(o.type, objpkg.KCLSchemaDefTypeObject) or isinstance(
            o.type, objpkg.KCLSchemaTypeObject
        ):
            if not schema_name or name == schema_name:
                result.append(kcl_type_obj_to_pb_kcl_type(o.type))
    return result


def kcl_type_obj_to_pb_kcl_type(tpe: types.Type) -> pb2.KclType:
    """any, schema, dict, list, str, int, float, bool, int(), float() str(), bool(), union"""
    if isinstance(tpe, objpkg.KCLSchemaDefTypeObject):
        return kcl_type_obj_to_pb_kcl_type(tpe.schema_type)
    elif isinstance(tpe, objpkg.KCLSchemaTypeObject):
        return schema_type_obj_to_pb_kcl_type(tpe)
    elif isinstance(tpe, objpkg.KCLAnyTypeObject):
        return pb2.KclType(
            type="any",
        )
    elif isinstance(tpe, objpkg.KCLDictTypeObject):
        return pb2.KclType(
            type="dict",
            key=kcl_type_obj_to_pb_kcl_type(tpe.key_type),
            item=kcl_type_obj_to_pb_kcl_type(tpe.value_type),
        )
    elif isinstance(tpe, objpkg.KCLListTypeObject):
        return pb2.KclType(
            type="list",
            item=kcl_type_obj_to_pb_kcl_type(tpe.item_type),
        )
    elif isinstance(
        tpe,
        (
            objpkg.KCLIntTypeObject,
            objpkg.KCLFloatTypeObject,
            objpkg.KCLStringTypeObject,
            objpkg.KCLBoolTypeObject,
            objpkg.KCLIntLitTypeObject,
            objpkg.KCLFloatLitTypeObject,
            objpkg.KCLStringLitTypeObject,
            objpkg.KCLBoolLitTypeObject,
        ),
    ):
        return pb2.KclType(
            type=tpe.type_str(),
        )
    elif isinstance(tpe, objpkg.KCLUnionTypeObject):
        return pb2.KclType(
            type="union",
            union_types=[kcl_type_obj_to_pb_kcl_type(t) for t in tpe.types],
        )
    return (
        pb2.KclType(
            type=tpe.type_str(),
        )
        if isinstance(tpe, types.Type)
        else None
    )


def schema_type_obj_to_pb_kcl_type(
    schema_type_obj: objpkg.KCLSchemaTypeObject,
) -> pb2.KclType:
    """Convert the schema type object to the protobuf kcl type."""
    return pb2.KclType(
        type="schema",
        schema_name=schema_type_obj.name,
        schema_doc=schema_type_obj.doc,
        properties=get_schema_type_obj_properties(schema_type_obj),
        decorators=[
            kcl_decorator_to_pb_decorator(decorator)
            for decorator in schema_type_obj.node_ref.decorators or []
        ],
        required=list(sorted(get_schema_type_obj_required_attributes(schema_type_obj))),
    )


def get_schema_type_obj_properties(
    schema_type_obj: objpkg.KCLSchemaTypeObject,
) -> Dict[str, pb2.KclType]:
    """Get schema properties from a schema type object"""
    result_map = {}
    base_result_map = (
        get_schema_type_obj_properties(schema_type_obj.base)
        if schema_type_obj.base
        else {}
    )
    for attr, attr_obj in schema_type_obj.attr_obj_map.items():
        if attr != objpkg.SCHEMA_SETTINGS_ATTR_NAME:
            type_node = attr_obj.attr_type
            result_map[attr] = kcl_type_obj_to_pb_kcl_type(type_node)
            result_map[attr].default = (
                value_to_string(attr_obj.attr_node.value)
                if attr_obj.attr_node.value
                else ""
            )
            result_map[attr].decorators.extend(
                [
                    kcl_decorator_to_pb_decorator(decorator)
                    for decorator in attr_obj.attr_node.decorators
                ]
                if isinstance(attr_obj.attr_node, ast.SchemaAttr)
                else []
            )
    base_result_map.update(result_map)
    line = 1
    for k in base_result_map:
        base_result_map[k].line = line
        line += 1
    return base_result_map


def get_schema_type_obj_required_attributes(
    schema_type_obj: objpkg.KCLSchemaTypeObject,
) -> Set[str]:
    """Get the required attributes from the schema type object"""
    base_attribute_set = (
        get_schema_type_obj_required_attributes(schema_type_obj.base)
        if schema_type_obj.base
        else set()
    )
    required_attribute_set = {
        attr
        for attr, attr_obj in schema_type_obj.attr_obj_map.items()
        if attr != objpkg.SCHEMA_SETTINGS_ATTR_NAME and not attr_obj.is_optional
    }
    required_attribute_set.update(base_attribute_set)
    return required_attribute_set


def kcl_decorator_to_pb_decorator(node: ast.Decorator) -> pb2.Decorator:
    """Convert the decorator node to the protobuf decorator type."""

    return pb2.Decorator(
        name=node.name.get_name(),
        arguments=[value_to_string(arg) for arg in node.args.args or [] if arg]
        if node.args
        else [],
        keywords={
            keyword.arg.get_name(): value_to_string(keyword.value)
            for keyword in node.args.keywords or []
            if keyword
        }
        if node.args
        else {},
    )


def value_to_string(node: ast.AST) -> str:
    """AST value to string"""
    if isinstance(node, ast.StringLit):
        return node.value
    else:
        buffer = io.StringIO()
        printer.PrintAST(node, out=buffer)
        return buffer.getvalue()



# Copyright 2020 The KCL Authors. All rights reserved.

import google.protobuf.json_format as json_format


def FromJson(text: str, message):
    return json_format.Parse(text, message)


def ToJson(message) -> str:
    return json_format.MessageToJson(
        message, including_default_value_fields=True, preserving_proto_field_name=True
    )



# Copyright 2020 The KCL Authors. All rights reserved.

from typing import Tuple, Union, List

import kclvm.kcl.ast as ast

# Auto Generated
AST_FIELDS_MAP = {
    # Stmt
    "ExprStmt": ["exprs"],
    "UnificationStmt": ["target", "value"],
    "TypeAliasStmt": ["type_name"],
    "AssignStmt": ["targets", "value", "type_annotation_node"],
    "AugAssignStmt": ["target", "op", "value"],
    "AssertStmt": ["test", "if_cond", "msg"],
    "IfStmt": ["cond", "body", "elif_cond", "elif_body", "else_body"],
    "ImportStmt": ["path", "name", "asname", "rawpath", "path_nodes", "as_name_node"],
    "SchemaAttr": [
        "doc",
        "name",
        "type_str",
        "op",
        "value",
        "is_optional",
        "decorators",
        "name_node",
        "type_node",
    ],
    "SchemaStmt": [
        "doc",
        "name",
        "parent_name",
        "is_mixin",
        "args",
        "mixins",
        "body",
        "decorators",
        "checks",
        "index_signature",
        "name_node",
        "for_host_name",
    ],
    "RuleStmt": [
        "doc",
        "name",
        "parent_rules",
        "decorators",
        "args",
        "checks",
        "name_node",
        "for_host_name",
    ],
    # Expr
    "QuantExpr": ["target", "variables", "test", "if_cond"],
    "SchemaIndexSignature": ["value", "name_node", "value_type_node"],
    "IfExpr": ["body", "cond", "orelse"],
    "UnaryExpr": ["op", "operand"],
    "BinaryExpr": ["left", "op", "right"],
    "SelectorExpr": ["value", "attr", "has_question"],
    "CallExpr": ["func", "args", "keywords"],
    "Subscript": ["value", "index", "lower", "upper", "step", "has_question"],
    "ParenExpr": ["expr"],
    "Operand": ["value"],
    "ListExpr": ["elts"],
    "ListComp": ["elt", "generators"],
    "ListIfItemExpr": ["if_cond", "exprs", "orelse"],
    "StarredExpr": ["value"],
    "DictComp": ["key", "value", "generators"],
    "ConfigIfEntryExpr": ["if_cond", "keys", "values", "orelse"],
    "CompClause": ["targets", "iter", "ifs"],
    "SchemaExpr": ["name", "args", "kwargs", "config"],
    "ConfigExpr": ["items"],
    "ConfigEntry": ["key", "value", "operation", "insert_index"],
    "CheckExpr": ["test", "if_cond", "msg"],
    "LambdaExpr": ["args", "body"],
    "Decorator": ["name", "args"],
    "Keyword": ["arg", "value"],
    "Arguments": ["args", "defaults", "type_annotation_node_list"],
    "Compare": ["left", "ops", "comparators"],
    "Identifier": ["names", "name_nodes"],
    "Name": ["value"],
    "Literal": ["value"],
    "NumberLit": ["value"],
    "StringLit": ["value", "is_long_string"],
    "NameConstantLit": ["value"],
    "JoinedString": ["values", "is_long_string"],
    "FormattedValue": ["value", "is_long_string", "format_spec"],
    "Comment": ["text"],
    "CommentGroup": ["comments"],
    "Module": ["body", "comments"],
    # Type
    "Type": ["type_elements"],
    "BasicType": ["type_name"],
    "ListType": ["inner_type"],
    "DictType": ["key_type", "value_type"],
    "LiteralType": ["string_value", "number_value"],
}


def iter_fields(t: ast.AST) -> List[Tuple[str, Union[ast.AST, List[ast.AST]]]]:
    """Return ast node attribute-value AST pair"""
    if not t or not isinstance(t, ast.AST):
        return []
    return [(field, getattr(t, field)) for field in AST_FIELDS_MAP[t.type]]



# Copyright 2021 The KCL Authors. All rights reserved.

import os
import typing
import json

from ast import literal_eval

import kclvm.config
import kclvm.kcl.ast as ast
import kclvm.api.object as objpkg
import kclvm.tools.query as query
import kclvm.compiler.parser.parser as parser
import kclvm.compiler.build.compiler as compiler
import kclvm.compiler.vfs as vfs
import kclvm.vm as vm
import kcl_lib.api.spec_pb2 as pb2

from kclvm.api.object.internal import (
    kcl_option_reset,
    kcl_option_init_all,
)
from .kclvm_cli import kclvm_cli_native_run_dylib


KCLVM_PANIC_INFO_KEY = "__kcl_PanicInfo__"
KCLVM_RUN_MODE_WITHIN_CACHE_ENV = "KCLVM_RUN_MODE_WITHIN_CACHE"
KCLVM_TARGET_ENV_KEY = "KCLVM_TARGET"


def Run(
    path_list: typing.List[str],
    *,
    work_dir: str = "",
    k_code_list: typing.List[str] = None,
    cmd_args: typing.List[ast.CmdArgSpec] = None,
    cmd_overrides: typing.List[ast.CmdOverrideSpec] = None,
    # -r --strict-range-check
    strict_range_check: bool = None,
    # -n --disable-none
    disable_none: bool = None,
    # -v --verbose
    verbose: int = None,
    # -d --debug
    debug: int = None,
    print_override_ast: bool = False,
    # --target
    target: str = "native",
) -> objpkg.KCLResult:
    assert len(path_list) > 0

    if not work_dir and not k_code_list:
        for s in path_list:
            if os.path.isdir(s):
                work_dir = s
    if not work_dir and not k_code_list:
        work_dir = kclvm.config.current_path or os.path.dirname(path_list[0])

    root = vfs.MustGetPkgRoot(path_list)
    modfile = vfs.LoadModFile(root)
    target = (
        target or modfile.build.target or os.getenv(KCLVM_TARGET_ENV_KEY) or ""
    ).lower()

    kclvm.config.input_file = path_list
    kclvm.config.current_path = work_dir
    kclvm.config.is_target_native = target == "native"

    if strict_range_check is not None:
        kclvm.config.strict_range_check = strict_range_check
    if disable_none is not None:
        kclvm.config.disable_none = disable_none
    if verbose is not None:
        kclvm.config.verbose = verbose
    if debug is not None:
        kclvm.config.debug = debug

    if cmd_args:
        kclvm.config.arguments = []
        for x in cmd_args or []:
            try:
                better_value = literal_eval(x.value)
                kclvm.config.arguments.append((x.name, better_value))
            except Exception:
                kclvm.config.arguments.append((x.name, x.value))

    # rust: build/link/run
    if target == "native":
        kclvm.config.is_target_native = True

        args = pb2.ExecProgram_Args()
        args.work_dir = work_dir
        args.k_filename_list.extend(path_list or [])
        args.k_code_list.extend(k_code_list or [])

        for kv in kclvm.config.arguments or []:
            key, value = kv
            if isinstance(value, (bool, list, dict)):
                value = json.dumps(value)
            elif isinstance(value, str):
                value = '"{}"'.format(value.replace('"', '\\"'))
            else:
                value = str(value)
            args.args.append(
                pb2.CmdArgSpec(
                    name=key,
                    value=value,
                )
            )

        for x in cmd_overrides or []:
            args.overrides.append(
                pb2.CmdOverrideSpec(
                    pkgpath=x.pkgpath,
                    field_path=x.field_path,
                    field_value=x.field_value,
                    action=x.action.value,
                )
            )

        args.print_override_ast = print_override_ast or False
        args.strict_range_check = strict_range_check or False
        args.disable_none = disable_none or False
        args.verbose = verbose or 0
        args.debug = debug or 0

        return kclvm_cli_native_run_dylib(args)

    ast_prog = parser.LoadProgram(
        *path_list,
        work_dir=work_dir,
        k_code_list=k_code_list,
        mode=parser.ParseMode.ParseComments if cmd_overrides else parser.ParseMode.Null,
    )
    ast_prog.cmd_args = cmd_args if cmd_args else []
    ast_prog.cmd_overrides = cmd_overrides if cmd_overrides else []

    # Apply argument
    kcl_option_reset()
    kcl_option_init_all()

    # AST to bytecode list
    bin_prog = compiler.CompileProgram(
        ast_prog, enable_cache=not bool(ast_prog.cmd_overrides)
    )

    # Run bytecode list
    result = vm.Run(bin_prog)

    # If cmd overrides are used and config.debug is True, write back KCL files
    if print_override_ast:
        query.PrintOverridesAST()
    return result



"""The `type_parser` file mainly contains the function `parse_type_str`
which is used to parser a type string to a KCL type object

:copyright: Copyright 2020 The KCL Authors. All rights reserved.
"""

import re as _re

from typing import Dict
from ast import literal_eval

import kclvm.api.object as objpkg
import kclvm.api.object.internal.common as common
import kclvm.compiler.build.utils.units as units

from .type import (
    Type,
    INT_TYPE,
    FLOAT_TYPE,
    STR_TYPE,
    BOOL_TYPE,
    ANY_TYPE,
    TRUE_LIT_TYPE,
    FALSE_LIT_TYPE,
    DICT_STR_ANY_TYPE,
    DICT_STR_STR_TYPE,
    sup,
)

BUILTIN_TYPES = ["str", "bool", "int", "float"]
_KCL_TYPE_any = "any"
_KCL_TYPE_True = "True"
_KCL_TYPE_False = "False"


TYPES_MAPPING: Dict[str, Type] = {
    "int": INT_TYPE,
    "float": FLOAT_TYPE,
    "str": STR_TYPE,
    "bool": BOOL_TYPE,
    "any": ANY_TYPE,
    "[]": objpkg.KCLListTypeObject(item_type=ANY_TYPE),
    "[any]": objpkg.KCLListTypeObject(item_type=ANY_TYPE),
    "[str]": objpkg.KCLListTypeObject(item_type=STR_TYPE),
    "{:}": objpkg.KCLDictTypeObject(key_type=ANY_TYPE, value_type=ANY_TYPE),
    "{str:}": DICT_STR_ANY_TYPE,
    "{str:any}": DICT_STR_ANY_TYPE,
    "{str:str}": DICT_STR_STR_TYPE,
}


def parse_type_str(tpe_str: str) -> Type:
    """Parser a type string to a type object"""
    if tpe_str is None or tpe_str == "":
        return ANY_TYPE
    if not isinstance(tpe_str, str):
        raise ValueError(f"Argument tpe_str must be str, not {type(tpe_str)}")
    # Remove the space in the type string
    tpe_str = tpe_str.strip(" \t\f\v\r\n")
    if tpe_str in TYPES_MAPPING:
        return TYPES_MAPPING[tpe_str]
    # Union type
    if is_union_type_str(tpe_str):
        return parse_union_type_str(tpe_str)
    # Bultin literal type
    elif is_lit_type_str(tpe_str):
        return parse_lit_type_str(tpe_str)
    # Number multiplier literal type
    elif is_number_multiplier_literal_type(tpe_str):
        return parse_number_multiplier_literal_type(tpe_str)
    # Dict type
    elif is_dict_type_str(tpe_str):
        k_type_str, v_type_str = common.separate_kv(common.dereferencetype(tpe_str))
        return objpkg.KCLDictTypeObject(
            key_type=parse_type_str(k_type_str),
            value_type=parse_type_str(v_type_str),
        )
    # List type
    elif is_list_type_str(tpe_str):
        return objpkg.KCLListTypeObject(
            item_type=parse_type_str(common.dereferencetype(tpe_str))
        )
    # Schema type or pkg.Schema type or named type
    return parse_named_type(tpe_str)


# -----------------------
# Judge type string
# -----------------------


def is_union_type_str(tpe: str) -> bool:
    """Whether a type string is a union type string, e.g. A|B|C,
    and detect '|' in type string except '|' in dict or list.
    """
    return common.is_type_union(tpe)


def is_list_type_str(tpe: str) -> bool:
    """Whether a type string is a list type string"""
    return common.islisttype(tpe)


def is_dict_type_str(tpe: str) -> bool:
    """Whether a type string is a dict type string"""
    return common.isdicttype(tpe)


def is_lit_type_str(tpe: str) -> bool:
    """Whether a type string is a literal type string"""
    if tpe in [_KCL_TYPE_True, _KCL_TYPE_False]:
        return True

    # str
    if tpe.startswith('"'):
        return tpe.endswith('"')
    if tpe.startswith("'"):
        return tpe.endswith("'")

    # int or float
    try:
        float(tpe)
        return True
    except ValueError:
        pass

    # non literal type
    return False


def is_number_multiplier_literal_type(tpe: str) -> bool:
    """Whether a type string is a number multiplier literal type string"""
    return bool(_re.match(units.NUMBER_MULTIPLIER_REGEX, tpe))


# -----------------------
# Parse type string
# -----------------------


def parse_union_type_str(tpe_str: str) -> Type:
    """Parse union type string"""
    type_str_list = common.split_type_union(tpe_str)
    types = [parse_type_str(tpe) for tpe in type_str_list]
    return sup(types)


def parse_lit_type_str(tpe_str: str) -> Type:
    """Parse literal type string"""
    type_val = literal_eval(tpe_str)
    if isinstance(type_val, bool):
        return TRUE_LIT_TYPE if type_val else FALSE_LIT_TYPE
    if isinstance(type_val, str):
        return objpkg.KCLStringLitTypeObject(value=type_val)
    if isinstance(type_val, int):
        return objpkg.KCLIntLitTypeObject(value=type_val)
    if isinstance(type_val, float):
        return objpkg.KCLFloatLitTypeObject(value=type_val)
    raise ValueError(f"Invalid argument tpe_str {tpe_str}")


def parse_number_multiplier_literal_type(tpe_str: str) -> Type:
    """Parse number multiplier literal type"""
    if tpe_str[-1] == units.IEC_SUFFIX:
        value, suffix = int(tpe_str[:-2]), tpe_str[-2:]
    else:
        value, suffix = int(tpe_str[:-1]), tpe_str[-1]
    return objpkg.KCLNumberMultiplierTypeObject(
        value=units.cal_num(value, suffix),
        raw_value=value,
        binary_suffix=suffix,
    )


def parse_named_type(tpe_str: str) -> Type:
    # Please note Named type to find it in the scope (e.g. schema type, type alias)
    return objpkg.KCLNamedTypeObject(name=tpe_str)


# -----------------------
# END
# -----------------------



""" The `code` module mainly contains the definition of KCL
bytecode. The bytecode factory is used for display debugging
in the `code_factory` file and the corresponding bytecode
execution function is in the `code_actions` file.

Each bytecode corresponds to an execution function. For example,
a binary addition operation corresponds to `Opcode.BINARY_ADD`.

KCL Bytecodes are mainly divided into three categories:
* Stack operation bytecodes
* KCL semantic related bytecode
* Internal bytecodes for debugging

:copyright: Copyright 2020 The KCL Authors. All rights reserved.
"""

from .code import (
    Opcode,
    Label,
    JumpAbs,
    JumpRel,
    Instruction,
    InstructionWithArg,
    EmittedInstruction,
    CompilationScope,
)
from .code_factory import SchemaBodyOpcodeFactory
from .code_actions import VM_OP_ACTIONS

__all__ = [
    "Opcode",
    "Label",
    "Instruction",
    "InstructionWithArg",
    "JumpAbs",
    "JumpRel",
    "EmittedInstruction",
    "CompilationScope",
    "SchemaBodyOpcodeFactory",
    "VM_OP_ACTIONS",
]



# Copyright 2020 The KCL Authors. All rights reserved.

import os
import glob
import re
import copy
import pathlib
import typing
import sys
import traceback
import inspect
import platform

import kclvm.kcl.info as kcl_info
import kclvm.compiler.extension.plugin.template as plugin_template


UNKNOWN_VERSION = "unknown"


class Init:
    def __init__(self, root: str = "", info_map=None):
        if info_map is None:
            info_map = {}
        self.plugins_root = root or ""
        self.plugins_info_map = info_map or {}


_re_plugin_name = re.compile("^[a-z][a-z0-9_-]*$")


def _is_valid_plugin_name(plugin_name: str) -> bool:
    return _re_plugin_name.match(plugin_name) is not None


def _normalize_plugin_name(plugin_name: str) -> str:
    if plugin_name.startswith("kcl_plugin.") or plugin_name.startswith(
        f"kcl_plugin{os.sep}"
    ):
        plugin_name = plugin_name[len("kcl_plugin.") :]
    return plugin_name


def _get_plugin(root: str, plugin_name: str) -> typing.Optional[any]:
    if not os.path.exists(f"{root}{os.sep}{plugin_name}{os.sep}plugin.py"):
        return None

    import importlib.util

    spec = importlib.util.spec_from_file_location(
        f"kcl_plugin.{plugin_name}", f"{root}{os.sep}{plugin_name}{os.sep}plugin.py"
    )
    pkg = importlib.util.module_from_spec(spec)

    try:
        spec.loader.exec_module(pkg)
    except Exception:
        ex_type, ex_val, ex_stack = sys.exc_info()
        print(
            f"WARN: {root}{os.sep}{plugin_name}{os.sep}plugin.py:{traceback.extract_tb(ex_stack)[-1].lineno}: init_plugin failed: {ex_val}"
        )
        return None

    func_map = {}
    for func_name, func_body in inspect.getmembers(pkg, inspect.isfunction):
        if func_name.startswith(kcl_info.MANGLE_PREFIX):
            func_map[func_name[len(kcl_info.MANGLE_PREFIX) :]] = func_body

    for func_name, func_body in inspect.getmembers(pkg, inspect.isfunction):
        if not func_name.startswith(kcl_info.MANGLE_PREFIX) and func_name[0] != "_":
            if func_name not in func_map:
                func_map[kcl_info.MANGLE_PREFIX + func_name] = func_body

    for func_name in func_map:
        setattr(pkg, func_name, func_map[func_name])

    return pkg


def _get_info(pkg) -> dict:
    info = copy.deepcopy(getattr(pkg, "INFO"))

    info["method"] = {}
    for s in dir(pkg):
        if s.startswith(kcl_info.MANGLE_PREFIX):
            func_name = s[len(kcl_info.MANGLE_PREFIX) :]
            func_doc = getattr(pkg, s).__doc__
            info["method"][func_name] = func_doc if func_doc else "no doc"

    return info


def find_plugin_root() -> typing.Optional[str]:
    return _find_plugin_root()


def _find_plugin_root() -> typing.Optional[str]:
    # 1. try $KCL_PLUGINS_ROOT env
    env_plugin_root = os.getenv("KCL_PLUGINS_ROOT", "")
    if env_plugin_root != "":
        return env_plugin_root

    # 2. try ${pwd}{os.sep}...{os.sep}plugins{os.sep}hello{os.sep}plugin.py
    cwd_plugin_path = pathlib.Path(
        os.environ.get("KCLVM_CLI_BIN_PATH") or os.getcwd()
    ).absolute()
    root = cwd_plugin_path.root
    while cwd_plugin_path:
        if cwd_plugin_path == cwd_plugin_path.parent or str(cwd_plugin_path) == root:
            break
        plugin_list_file_path = cwd_plugin_path.joinpath(
            f"plugins{os.sep}hello{os.sep}plugin.py"
        )
        if plugin_list_file_path.exists() and plugin_list_file_path.is_file():
            return str(cwd_plugin_path.joinpath("plugins"))
        if cwd_plugin_path.joinpath("kcl.mod").exists():
            break
        cwd_plugin_path = cwd_plugin_path.parent

    # 3. try ${__file__}{os.sep}...{os.sep}plugins{os.sep}hello{os.sep}plugin.py
    cwd_plugin_path = pathlib.Path(__file__).parent.absolute()
    root = cwd_plugin_path.root
    while cwd_plugin_path:
        if cwd_plugin_path == cwd_plugin_path.parent or str(cwd_plugin_path) == root:
            break
        plugin_list_file_path = cwd_plugin_path.joinpath(
            f"plugins{os.sep}hello{os.sep}plugin.py"
        )
        if plugin_list_file_path.exists() and plugin_list_file_path.is_file():
            return str(cwd_plugin_path.joinpath("plugins"))
        cwd_plugin_path = cwd_plugin_path.parent

    # 4. try $HOME{os.sep}.kcl{os.sep}plugins
    home_dir = (
        os.getenv("HOME")
        if platform.system() != "Windows"
        else os.getenv("UserProfile")
    )
    home_plugin_root = os.path.join(home_dir, f".kcl{os.sep}plugins")
    if os.path.exists(f"{home_plugin_root}{os.sep}hello{os.sep}plugin.py"):
        return home_plugin_root

    # 5. not found
    return None


def _init_plugin_root() -> typing.Tuple[typing.Optional[str], dict]:
    plugins_root = _find_plugin_root()
    if plugins_root is None:
        return None, {}

    plugins_info = {}

    # 'hello' is builtin plugin, and used in test code
    if not os.path.exists(f"{plugins_root}{os.sep}hello{os.sep}plugin.py"):
        os.makedirs(f"{plugins_root}{os.sep}hello")

        with open(f"{plugins_root}{os.sep}hello{os.sep}plugin.py", "w") as file:
            file.write(plugin_template.get_plugin_template_code("hello"))
        with open(f"{plugins_root}{os.sep}hello{os.sep}plugin_test.py", "w") as file:
            file.write(plugin_template.get_plugin_test_template_code("hello"))

    # scan all plugins
    k_files = glob.glob(f"{plugins_root}{os.sep}*{os.sep}plugin.py", recursive=False)
    for i in range(len(k_files)):
        plugin_name = os.path.basename(k_files[i][: -len(f"{os.sep}plugin.py")])
        if _is_valid_plugin_name(plugin_name):
            pkg = _get_plugin(plugins_root, plugin_name)
            if not pkg:
                continue
            info = _get_info(pkg)

            plugins_info[plugin_name] = info
    return plugins_root, plugins_info


# init plugins
_plugin_root, _plugin_info_map = _init_plugin_root()
init_ = Init(_plugin_root, _plugin_info_map)


# -----------------------------------------------------------------------------
# API
# -----------------------------------------------------------------------------


def reset_plugin(plugin_root: str = ""):
    global _plugin_root
    global _plugin_info_map
    global init_

    os.environ["KCL_PLUGINS_ROOT"] = f"{plugin_root}"
    _plugin_root, _plugin_info_map = _init_plugin_root()
    init_ = Init(_plugin_root, _plugin_info_map)


def get_plugin_version() -> str:
    if not init_.plugins_root:
        return UNKNOWN_VERSION
    version_path = pathlib.Path(f"{init_.plugins_root}{os.sep}VERSION")
    if version_path.exists():
        return version_path.read_text()
    return UNKNOWN_VERSION


def get_plugin_root(plugin_name: str = "") -> typing.Optional[str]:
    if init_.plugins_root is None:
        return None
    if plugin_name != "":
        plugin_name = _normalize_plugin_name(plugin_name)
        return f"{init_.plugins_root}{os.sep}{plugin_name}"

    return init_.plugins_root


def get_plugin_names() -> typing.List[str]:
    if init_.plugins_root is None:
        return []
    plugin_names = []
    for s in init_.plugins_info_map:
        plugin_names.append(s)

    plugin_names.sort()
    return plugin_names


def get_info(plugin_name: str) -> typing.Optional[dict]:
    if init_.plugins_root is None:
        return None

    plugin_name = _normalize_plugin_name(plugin_name)

    if plugin_name not in init_.plugins_info_map:
        return None
    return init_.plugins_info_map[plugin_name]


def get_source_code(plugin_name: str) -> typing.Optional[str]:
    if init_.plugins_root is None:
        return None

    plugin_name = _normalize_plugin_name(plugin_name)

    if plugin_name not in init_.plugins_info_map:
        return None

    code = ""
    with open(f"{init_.plugins_root}{os.sep}{plugin_name}{os.sep}plugin.py") as f:
        code += f.read()
    return code


def get_plugin(plugin_name: str) -> typing.Optional[any]:
    if init_.plugins_root is None:
        return None

    plugin_name = _normalize_plugin_name(plugin_name)
    return _get_plugin(init_.plugins_root, plugin_name)


# -----------------------------------------------------------------------------
# UTILS
# -----------------------------------------------------------------------------


def init_plugin(plugin_name: str):
    if init_.plugins_root is None:
        return None

    plugin_name = _normalize_plugin_name(plugin_name)

    if not _is_valid_plugin_name(plugin_name):
        print(f'WARN: init_plugin("{plugin_name}") failed, invalid name')
        return

    if os.path.exists(f"{init_.plugins_root}{os.sep}{plugin_name}{os.sep}plugin.py"):
        print(f'WARN: init_plugin("{plugin_name}") failed, plugin exists')
        return

    golden_plugin_skectch_code = plugin_template.get_plugin_template_code(plugin_name)
    golden_plugin_skectch_code_test = plugin_template.get_plugin_test_template_code(
        plugin_name
    )

    if not os.path.exists(f"{init_.plugins_root}{os.sep}{plugin_name}"):
        os.makedirs(f"{init_.plugins_root}{os.sep}{plugin_name}")

    with open(
        f"{init_.plugins_root}{os.sep}{plugin_name}{os.sep}plugin.py", "w"
    ) as file:
        file.write(golden_plugin_skectch_code)

    with open(
        f"{init_.plugins_root}{os.sep}{plugin_name}{os.sep}plugin_test.py", "w"
    ) as file:
        file.write(golden_plugin_skectch_code_test)

    gendoc(plugin_name)


def gendoc(plugin_name: str):
    if init_.plugins_root is None:
        print("WARN: plugin root not found")
        return

    if not _is_valid_plugin_name(plugin_name):
        print(f'WARN: gendoc("{plugin_name}") failed, invalid name')
        return

    pkg = _get_plugin(init_.plugins_root, plugin_name)
    if not pkg:
        print("WARN: plugin init failed")
        return
    info = _get_info(pkg)
    if info is None:
        print(f'WARN: gendoc("{plugin_name}") failed, not found plugin')
        return

    with open(f"{init_.plugins_root}{os.sep}{plugin_name}{os.sep}api.md", "w") as file:
        file.write(f"# plugin: `{info['name']}` - {info['describe']}\n\n")
        file.write(f"{info['long_describe']}\n\n")
        file.write(f"*version: {info['version']}*\n\n")

        for func_name in info["method"]:
            func_doc = info["method"][func_name]

            func_doc_line = []
            for line in func_doc.splitlines():
                if line.startswith("    "):
                    line = line[4:]
                func_doc_line.append(line)

            file.write(f"## `{func_name}`\n\n")
            for line in func_doc_line:
                file.write(f"{line}\n")

            file.write("\n")

    return


# -----------------------------------------------------------------------------
# END
# -----------------------------------------------------------------------------



import ruamel.yaml as yaml
import json
import io

from abc import ABC, abstractmethod

import kclvm.tools.docs.pb as pb
import kclvm.tools.docs.model_pb2 as model
import kclvm.tools.docs.factory as factory
import kclvm.tools.docs.formats as doc_formats

# ---------------------------------------------------
# Constants
# ---------------------------------------------------


ENDLINE = "\n"


# -------------------------------------------------------
# Document reader from KCL ModuleDoc type ot markdown type
# -------------------------------------------------------


class ModuleDocReader(ABC):
    """Module document reader class.

    This class provides basic functions for reader KCL Module from the IO.
    """

    def __init__(self, io_in: io.TextIOBase):
        self.io_in: io.TextIOBase = io_in

    @abstractmethod
    def read_doc(self) -> model.ModuleDoc:
        pass

    def read(self) -> str:
        return self.io_in.read()


class JsonDocReader(ModuleDocReader):
    """JSON document reader based on ModuleDocReader.

    The class is used to read KCL module document from JSON file.
    """

    def _load_module_from_json(self, data) -> model.ModuleDoc:
        return pb.FromJson(data, model.ModuleDoc())

    def read_doc(self) -> model.ModuleDoc:
        json_data = self.read()
        return self._load_module_from_json(json_data)


class YamlDocReader(JsonDocReader):
    """YAML document reader based on JsonDocReader.

    The class is used to read KCL module document from YAML file
    """

    def read_doc(self) -> model.ModuleDoc:
        dict_data = yaml.safe_load(self.io_in)
        json_data = json.dumps(dict_data)
        return self._load_module_from_json(json_data)


class MarkDownDocReader(YamlDocReader):
    """Markdown document reader based on KCLModuleDocReader.

    The class is used to read KCL module document from markdown file.
    """

    def read_doc(self) -> model.ModuleDoc:
        """TODO: Parse a Markdown string to KCL module document model."""
        raise NotImplementedError()


factory = factory.IOFactory()
factory.register_format(doc_formats.KCLDocFormat.JSON, JsonDocReader)
factory.register_format(doc_formats.KCLDocFormat.YAML, YamlDocReader)
factory.register_format(doc_formats.KCLDocFormat.MARKDOWN, MarkDownDocReader)



# Copyright 2020 The KCL Authors. All rights reserved.


def get_plugin_template_code(plugin_name: str) -> str:
    return f'''# Copyright 2020 The KCL Authors. All rights reserved.

INFO = {{
    'name': '{plugin_name}',
    'describe': '{plugin_name} doc',
    'long_describe': 'long describe',
    'version': '0.0.1',
}}


global_int: int = 0


def set_global_int(v: int):
    global global_int
    global_int = v


def get_global_int() -> int:
    return global_int


def say_hello(msg: str):
    print('{plugin_name}.say_hello:', msg)
    return None


def add(a: int, b: int) -> int:
    """add two numbers, and return result"""
    return a + b


def tolower(s: str) -> str:
    return s.lower()


def update_dict(d: dict, key: str, value: str) -> dict:
    d[key] = value
    return d


def list_append(l: list, *values) -> list:
    for v in values:
        l.append(v)
    return l


def foo(a, b, *, x, **values):
    print(a, b, x, values)
    return {{'a': a, 'b': b, 'x': x, **values}}
'''


def get_plugin_test_template_code(plugin_name: str) -> str:
    return """# Copyright 2020 The KCL Authors. All rights reserved.

# python3 -m pytest

import plugin


def test_add():
    assert plugin.add(1, 2) == 3


def test_tolower():
    assert plugin.tolower('KCL') == 'kcl'


def test_update_dict():
    assert plugin.update_dict({{'name': 123}}, 'name', 'kcl')['name'] == 'kcl'


def test_list_append():
    l = plugin.list_append(['abc'], 'name', 123)
    assert len(l) == 3
    assert l[0] == 'abc'
    assert l[1] == 'name'
    assert l[2] == 123


def test_foo():
    v = plugin.foo('aaa', 'bbb', x=123, y=234, abcd=1234)
    assert len(v) == 5
    assert v['a'] == 'aaa'
    assert v['b'] == 'bbb'
    assert v['x'] == 123
    assert v['y'] == 234
    assert v['abcd'] == 1234

    v = plugin.foo('aaa', 'bbb', x=123)
    assert len(v) == 3
    assert v['a'] == 'aaa'
    assert v['b'] == 'bbb'
    assert v['x'] == 123
"""



# Copyright 2021 The KCL Authors. All rights reserved.

import io
from typing import List
from dataclasses import dataclass

import kclvm.kcl.ast as ast

from .token import TokenValue
from .parser import (
    Import,
    Package,
    Option,
    Field,
    OneOfField,
    OneOf,
    Map,
    Reserved,
    Range,
    EnumField,
    Enum,
    Message,
    Service,
    Rpc,
    Proto,
    ImportOption,
    Type,
    KeyType,
)


PROTO_NODE_TUPLE = (
    Import,
    Package,
    Option,
    Field,
    OneOfField,
    OneOf,
    Map,
    Reserved,
    Range,
    EnumField,
    Enum,
    Message,
    Service,
    Rpc,
    Proto,
    ImportOption,
    Type,
    KeyType,
)

_INVALID_NODE_MSG = "Invalid proto node"
WHITESPACE = " "
TAB = "\t"
NEWLINE = "\n"


# ---------------------------------------------------
# Printer config
# ---------------------------------------------------


@dataclass
class Config:
    tab_len: int = 4
    indent_len: int = 4
    use_spaces: bool = True


class BasePrinter(ast.TreeWalker):
    def __init__(self, config: Config, out: io.TextIOBase):
        self.output: str = ""
        self.config: Config = config
        self.out: io.TextIOBase = out
        self.indent: int = 0

    # Base walker functions

    def get_node_name(self, t):
        """Get the ast.AST node name"""
        return type(t).__name__

    def enter(self):
        self.indent += 1

    def leave(self):
        self.indent -= 1

    # ---------------
    # Write functions
    # ---------------

    @staticmethod
    def interleave(inter, f, seq):
        """Call f on each item in seq, calling inter() in between."""
        if not seq:
            return
        seq = iter(seq)
        f(next(seq))
        for x in seq:
            inter()
            f(x)

    def write(self, text: str = ""):
        self.out.write(text)

    def fill(self):
        self.write(
            (self.indent * self.config.indent_len * WHITESPACE)
            if self.config.use_spaces
            else (TAB * self.indent)
        )

    def print(self, *values):
        for value in values or []:
            if isinstance(value, PROTO_NODE_TUPLE):
                self.walk(value)
            elif value is True:
                self.write("true")
            elif value is False:
                self.write("false")
            else:
                self.write(str(value))


class Printer(BasePrinter):
    def __init__(self, config: Config, out: io.TextIOBase):
        super().__init__(config, out)

    def walk_Proto(self, t: Proto):
        self.print(
            TokenValue.SYNTAX,
            WHITESPACE,
            TokenValue.EQ,
            WHITESPACE,
            f'"{t.syntax}"',
            TokenValue.SEMI,
            NEWLINE,
        )
        for node in t.statements or []:
            self.walk(node)

    def walk_Import(self, t: Import):
        self.print(
            TokenValue.IMPORT,
            WHITESPACE,
            t.option.value,
            WHITESPACE,
            f'"{t.identifier}"',
            TokenValue.SEMI,
            NEWLINE,
        )

    def walk_Package(self, t: Package):
        self.print(
            TokenValue.PACKAGE,
            WHITESPACE,
        )
        self.interleave(
            lambda: self.print(TokenValue.COMMA), lambda n: self.print(n), t.identifier
        )
        self.print(
            TokenValue.SEMI,
            NEWLINE,
        )

    def walk_Option(self, t: Option):
        self.fill()
        self.print(
            TokenValue.OPTION,
            WHITESPACE,
            t.name,
            WHITESPACE,
            TokenValue.EQ,
            WHITESPACE,
            f'"{t.value}"' if isinstance(t.value, str) else t.value,
            TokenValue.SEMI,
            NEWLINE,
        )

    def walk_Enum(self, t: Enum):
        self.write(NEWLINE)
        self.fill()
        self.print(
            TokenValue.ENUM,
            WHITESPACE,
            t.name,
            WHITESPACE,
            TokenValue.LBRACE,
            NEWLINE,
        )
        self.enter()
        for node in t.body:
            self.walk(node)
        self.leave()
        self.fill()
        self.print(TokenValue.RBRACE)
        self.write(NEWLINE)

    def walk_EnumField(self, t: EnumField):
        self.fill()
        self.print(
            t.name,
            WHITESPACE,
            TokenValue.EQ,
            WHITESPACE,
            t.value,
        )
        self.write_field_options(t.options)
        self.write(TokenValue.SEMI)
        self.write(NEWLINE)

    def walk_Message(self, t: Message):
        self.write(NEWLINE)
        self.fill()
        self.print(
            TokenValue.MESSAGE,
            WHITESPACE,
            t.name,
            WHITESPACE,
            TokenValue.LBRACE,
            NEWLINE,
        )
        self.enter()
        for node in t.body:
            self.walk(node)
        self.leave()
        self.fill()
        self.print(TokenValue.RBRACE)
        self.write(NEWLINE)

    def walk_Map(self, t: Map):
        self.fill()
        self.print(
            TokenValue.MAP,
            TokenValue.LANGLE_BRACK,
            t.key_type,
            TokenValue.COMMA,
            WHITESPACE,
            t.type,
            TokenValue.RANGLE_BRACK,
            WHITESPACE,
            t.name,
            WHITESPACE,
            TokenValue.EQ,
            WHITESPACE,
            t.number,
        )
        self.write_field_options(t.options)
        self.write(TokenValue.SEMI)
        self.write(NEWLINE)

    def walk_Field(self, t: Field):
        self.fill()
        if t.repeated:
            self.print(
                TokenValue.REPEATED,
                WHITESPACE,
            )
        self.print(
            t.type, WHITESPACE, t.name, WHITESPACE, TokenValue.EQ, WHITESPACE, t.number
        )
        self.write_field_options(t.options)
        self.write(TokenValue.SEMI)
        self.write(NEWLINE)

    def walk_OneOf(self, t: OneOf):
        self.fill()
        self.print(
            TokenValue.ONEOF,
            WHITESPACE,
            t.name,
            WHITESPACE,
            TokenValue.LBRACE,
            NEWLINE,
        )
        self.enter()
        for node in t.fields:
            self.walk(node)
        self.leave()
        self.fill()
        self.print(TokenValue.RBRACE)
        self.write(NEWLINE)

    def walk_OneOfField(self, t: OneOfField):
        self.fill()
        self.print(
            t.type, WHITESPACE, t.name, WHITESPACE, TokenValue.EQ, WHITESPACE, t.number
        )
        self.write_field_options(t.options)
        self.write(TokenValue.SEMI)
        self.write(NEWLINE)

    def walk_Service(self, t: Service):
        self.write(NEWLINE)
        self.fill()
        self.print(
            TokenValue.SERVICE,
            WHITESPACE,
            t.name,
            WHITESPACE,
            TokenValue.LBRACE,
            NEWLINE,
        )
        self.enter()
        for node in t.body:
            self.walk(node)
        self.leave()
        self.fill()
        self.print(TokenValue.RBRACE)
        self.write(NEWLINE)

    def walk_Rpc(self, t: Rpc):
        self.fill()
        self.print(
            TokenValue.RPC,
            WHITESPACE,
            t.name,
            TokenValue.LPAREN,
            t.request_message_type,
            TokenValue.RPAREN,
            WHITESPACE,
            TokenValue.RETURNS,
            WHITESPACE,
            TokenValue.LPAREN,
            t.response_message_type,
            TokenValue.RPAREN,
        )
        self.write_field_options(t.options)
        self.write(TokenValue.SEMI)
        self.write(NEWLINE)

    def walk_Reserved(self, t: Reserved):
        self.fill()
        self.print(
            TokenValue().RESERVED,
            WHITESPACE,
        )
        self.interleave(
            lambda: self.write(TokenValue.COMMA + WHITESPACE),
            lambda n: self.print(f'"{n}"' if isinstance(n, str) else n),
            t.items,
        )
        self.write(TokenValue.SEMI)
        self.write(NEWLINE)

    def walk_Range(self, t: Range):
        self.print(f'"{t.from_}"' if isinstance(t.from_, str) else t.from_)
        if t.to:
            self.print(
                WHITESPACE,
                TokenValue.TO,
                WHITESPACE,
                f'"{t.to}"' if isinstance(t.to, str) else t.to,
            )

    def walk_Type(self, t: Type):
        self.print(t.value)

    def walk_KeyType(self, t: KeyType):
        self.print(t.value)

    def write_field_options(self, options: List[Option]):
        def write_option(option: Option):
            self.print(
                TokenValue.LPAREN,
                option.name,
                TokenValue.RPAREN,
                WHITESPACE,
                TokenValue.EQ,
                WHITESPACE,
                f'"{option.value}"' if isinstance(option.value, str) else option.value,
            )

        if not options:
            return
        self.write(WHITESPACE)
        self.write(TokenValue.LBRACK)
        self.interleave(
            lambda: self.write(TokenValue.COMMA + WHITESPACE), write_option, options
        )
        self.write(TokenValue.RBRACK)


def print_node_to_string(
    node,
    config: Config = Config(),
) -> str:
    """Print a proto node to string io with `config`"""
    out = io.StringIO()
    Printer(config, out).walk(node)
    return out.getvalue()



from .info import (
    INT32_MIN,
    INT32_MAX,
    INT64_MIN,
    INT64_MAX,
    FLOAT32_MAX,
    FLOAT32_MIN,
    FLOAT32_EPSILON,
    FLOAT64_MAX,
    FLOAT64_MIN,
    FLOAT64_EPSILON,
    KCL_FILE_PATTERN,
    KCL_FILE_SUFFIX,
)
from .naming import (
    MANGLE_PREFIX,
    TAGGING_PREFIX,
    mangle,
    demangle,
    ismangled,
    tagging,
    detagging,
    istagged,
    isprivate_field,
)

__all__ = [
    "INT32_MIN",
    "INT32_MAX",
    "INT64_MIN",
    "INT64_MAX",
    "FLOAT32_MAX",
    "FLOAT32_MIN",
    "FLOAT32_EPSILON",
    "FLOAT64_MAX",
    "FLOAT64_MIN",
    "FLOAT64_EPSILON",
    "KCL_FILE_PATTERN",
    "KCL_FILE_SUFFIX",
    "MANGLE_PREFIX",
    "TAGGING_PREFIX",
    "mangle",
    "demangle",
    "ismangled",
    "tagging",
    "detagging",
    "istagged",
    "isprivate_field",
]



"""
The `check_utils` file mainly contains some methods for defensive programming.

Method `PreCheck` can be used for pre-checking the method,
mainly to verify whether the incoming parameters of the method meet conditions.

Method `PostCheck` can be used for post-checking the method,
mainly to verify whether the return of the method meets conditions.

Method `PostSimpleExprCheck` can be used for post-checking the method,
Compared with `PostCheck`, `PostSimpleExprCheck` supports verifying
some simple relationship between the return value and the input parameters

For example：

# Check whether the type of incoming parameter "a" is int
@PreCheck((lambda v: isinstance(v, int)), "a")
# Check whether the type of incoming parameter "b" is int
@PreCheck((lambda v: isinstance(v, int)), "b")
# Check whether the type of return value is int
@PostCheck((lambda v: isinstance(v, int)))
# Check whether the return value is equal to the sum of input parameters
@PostSimpleExprCheck((lambda inputs, result: result == inputs["a"] + inputs["b"]), ["a", "b"])
def add(a, b):
    return a + b

Class `CheckRules` and some other global methods are built-in check rules
that can be used in `PreCheck` and `PostCheck`.

:copyright: Copyright 2020 The KCL Authors. All rights reserved.
"""
import locale
import typing
import functools
from inspect import Signature
from typing import Callable, Any, List

# CHECK_MODE is a switch,
# you can return the parameters directly by turning the CHECK_MODE = False

CHECK_MODE = False


class CheckRules:
    @staticmethod
    def check_list_len_equal(all_lists: list):
        if not CHECK_MODE:
            return
        assert all_lists
        assert isinstance(all_lists, list)
        assert all(
            result is True
            for result in [isinstance(list_inner, list) for list_inner in all_lists]
        )
        assert all(
            result == len(all_lists[0])
            for result in [len(list_inner) for list_inner in all_lists]
        )

    @staticmethod
    def check_locale(lang: str):
        if not CHECK_MODE:
            return True
        LOCALE_LIST = list(locale.locale_alias.keys())
        if not lang or not isinstance(lang, str) or lang not in LOCALE_LIST:
            return False
        return True

    @staticmethod
    def check_type_not_none(item, *tpes) -> bool:
        if CHECK_MODE:
            return item is not None and isinstance(item, tpes)
        return True

    @staticmethod
    def check_type_allow_none(item, *tpes) -> bool:
        if CHECK_MODE:
            return item is None or isinstance(item, tpes)
        return True

    @staticmethod
    def check_list_item_type_allow_none(item, *tpes) -> bool:
        if CHECK_MODE:
            check_all_allow_none(list, item, *tpes)
        return True

    @staticmethod
    def check_int_range_allow_none(target: int, low: int, high: int) -> bool:
        if CHECK_MODE:
            if target is None:
                return True
            check_type_not_none(target, int)
            check_type_not_none(low, int)
            check_type_not_none(high, int)
            return target in range(low, high)
        else:
            return True

    @staticmethod
    def check_str_len_not_none(target: str, length: int) -> bool:
        if CHECK_MODE:
            check_type_not_none(target, str)
            check_type_not_none(length, int)
            return len(target) == length
        else:
            return True

    @staticmethod
    def check_str_len_allow_none(target: str, length: int) -> bool:
        if CHECK_MODE:
            if target is None:
                return True
            check_type_not_none(target, str)
            check_type_not_none(length, int)
            return len(target) == length
        else:
            return True


def check_allow_none(node, tpe):
    if node and CHECK_MODE:
        assert isinstance(node, tpe)
    return typing.cast(tpe, node)


def check_all_allow_none(set_tpe: typing.Type, nodes, *item_tpes):
    if nodes and CHECK_MODE:
        assert isinstance(nodes, set_tpe)
        assert isinstance(nodes, (list, tuple)) and all(
            isinstance(item, item_tpes) for item in nodes
        )
    return typing.cast(set_tpe, nodes)


def check_not_none(node, *tpes):
    if CHECK_MODE:
        assert node and isinstance(node, tpes)
    return typing.cast(tpes, node)


def check_all_not_none(set_tpe: typing.Type, nodes, *item_tpes):
    if CHECK_MODE:
        assert nodes and isinstance(nodes, set_tpe)
        assert isinstance(nodes, (list, tuple)) and all(
            isinstance(item, item_tpes) for item in nodes
        )
    return typing.cast(set_tpe, nodes)


def check_type_allow_none(node, *tpes):
    if node and CHECK_MODE:
        assert isinstance(node, tpes)
    return node


def check_type_not_none(node, *tpes):
    if CHECK_MODE:
        assert node is not None and isinstance(node, tpes)
    return node


def alert_internal_bug():
    if CHECK_MODE:
        assert False, "Here is unreachable unless a bug occurs"


def PreCheck(condition: Callable[[Any], bool], param_name: str, param_pos: int = None):
    def conditioner(func):
        @functools.wraps(func)
        def check_condition(*args, **kwargs):
            if not CHECK_MODE:
                return func(*args, **kwargs)
            check_type_not_none(condition, Callable)
            check_type_not_none(func, Callable)
            check_type_not_none(CHECK_MODE, bool)
            check_type_not_none(param_name, str)
            check_type_allow_none(param_pos, int)
            param_names_list = [
                i[0]
                for i in Signature.from_callable(func).parameters.items()
                if len(i) > 0
            ]
            if param_name not in param_names_list:
                raise AssertionError(
                    f"Pre-Condition failed: "
                    f"There is no parameter named {param_name} in function {func.__name__}. "
                    f"The function parameters list: {param_names_list}."
                )
            try:
                param_value = kwargs[
                    param_name
                ]  # if the param in kwargs for the function
            except KeyError:
                if not CheckRules.check_int_range_allow_none(
                    param_pos, 0, len(param_names_list)
                ):
                    raise AssertionError(
                        f"Pre-Condition failed: param_pos: {param_pos} is out of range. "
                        f"There are only {len(param_names_list)} parameters in {func.__name__}"
                    )
                if param_names_list.index(param_name) < len(
                    args
                ):  # if the param in args for the function
                    param_value = (
                        args[param_pos]
                        if param_pos
                        else args[param_names_list.index(param_name)]
                    )
                else:
                    param_value = None

            if condition(param_value):
                return func(*args, **kwargs)
            else:
                raise AssertionError(
                    f"Pre-Condition failed: {func.__name__}({param_name} = {param_value}), "
                    f"Check Condition: {condition.__name__}"
                )

        return check_condition

    return conditioner


def PostCheck(condition: Callable[[Any], bool]):
    def conditioner(func):
        @functools.wraps(func)
        def check_condition(*args, **kwargs):
            if not CHECK_MODE:
                return func(*args, **kwargs)
            check_type_not_none(condition, Callable)
            check_type_not_none(func, Callable)
            check_type_not_none(CHECK_MODE, bool)
            result = func(*args, **kwargs)
            if condition(result):
                return result
            else:
                raise AssertionError(
                    f"Post-Condition failed: {func.__name__} with returned {result},"
                    f"Check Condition: {condition.__name__}"
                )

        return check_condition

    return conditioner


def PostSimpleExprCheck(
    condition: Callable[[Any, Any], bool], dependent_params: List[str]
):
    def conditioner(func):
        @functools.wraps(func)
        def check_condition(*args, **kwargs):
            if not CHECK_MODE:
                return func(*args, **kwargs)
            check_type_not_none(condition, Callable)
            check_type_not_none(func, Callable)
            check_type_not_none(CHECK_MODE, bool)
            check_all_not_none(list, dependent_params, str)
            sig_params = {}
            param_names_list = [
                i[0]
                for i in Signature.from_callable(func).parameters.items()
                if len(i) > 0
            ]
            for k, v in Signature.from_callable(func).parameters.items():
                sig_params[k] = (k, v)
            inputs = {}

            for param in dependent_params:
                try:
                    inputs[param] = kwargs[
                        param
                    ]  # if the param in kwargs for the function
                except KeyError:
                    try:
                        inputs[param] = args[  # if the param in args for the function
                            param_names_list.index(sig_params[param][0])
                        ]
                    except IndexError:
                        assert (
                            len(sig_params[param]) > 0
                        )  # if the param in the function has default value
                        inputs[param] = sig_params[param][1].default
                    except KeyError:
                        raise AssertionError(
                            f"Post-Condition failed: "
                            f"There is no parameter named {param} in function {func.__name__}. "
                            f"The function parameters list: {param_names_list}."
                        )

            result = func(*args, **kwargs)
            if condition(inputs, result):
                return result
            else:
                raise AssertionError(
                    f"Post-condition failed: {func.__name__} with inputs: {inputs}, returned {result}"
                )

        return check_condition

    return conditioner



# Copyright 2021 The KCL Authors. All rights reserved.

import argparse
import sys
import pathlib

import kclvm.kcl.error as kcl_error

from .validation import validate_code


class ValidationMeta:
    DATA_FILE_DESC = "Validation data file"
    KCL_FILE_DESC = "KCL file"
    FORMAT_DESC = "Validation data file format, support YAML and JSON"


def kcl_vet_main():
    """KCL validation CLI main function."""
    parser = argparse.ArgumentParser(prog="kcl-vet")
    parser.add_argument(
        "-d",
        "--schema",
        dest="schema",
        metavar="schema",
        type=str,
        default=None,
        required=False,
    )
    parser.add_argument(
        "--format",
        dest="format",
        metavar="format",
        type=str,
        default="json",
        required=False,
        help=ValidationMeta.FORMAT_DESC,
    )
    parser.add_argument(
        "-n",
        "--attribute-name",
        dest="attribute_name",
        metavar="attribute_name",
        type=str,
        default="value",
        required=False,
    )
    parser.add_argument(
        dest="data_file",
        metavar="data_file",
        type=str,
        help=ValidationMeta.DATA_FILE_DESC,
    )
    parser.add_argument(
        dest="kcl_file",
        metavar="kcl_file",
        type=str,
        help=ValidationMeta.KCL_FILE_DESC,
    )
    args = parser.parse_args()
    if len(sys.argv) == 1:
        parser.print_help(sys.stdout)
        sys.exit(0)

    data_str = pathlib.Path(args.data_file).read_text()
    code_str = pathlib.Path(args.kcl_file).read_text()

    # Validate code
    if validate_code(
        data_str, code_str, args.schema, args.attribute_name, args.format, args.kcl_file
    ):
        print("Validate succuss!")


if __name__ == "__main__":
    try:
        kcl_vet_main()
    except kcl_error.KCLException as err:
        kcl_error.print_kcl_error_message(err, file=sys.stderr)
        sys.exit(1)
    except OSError as err:
        kcl_error.print_common_error_message(err, file=sys.stderr)
        sys.exit(1)
    except AssertionError as err:
        kcl_error.print_internal_error_message(err, file=sys.stderr)
        raise
    except Exception:
        kcl_error.print_internal_error_message(file=sys.stderr)
        raise
